# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from typing import Optional

import torch

from nemo.collections.asr.parts.utils.asr_confidence_utils import ConfidenceMethodMixin
from nemo.collections.asr.parts.utils.batched_beam_decoding_utils import (
    INACTIVE_SCORE,
    NON_EXISTENT_LABEL_VALUE,
    BatchedBeamHyps,
    BlankLMScoreMode,
    PruningMode,
)
from nemo.utils import logging


class ModifiedAESBatchedRNNTComputer(ConfidenceMethodMixin):
    """
    Batched Adaptive Expansion search implementation. Callable.
    Based on https://ieeexplore.ieee.org/document/9250505 with the following modficiations:
        - does not support prediction network caching
        - supports prefix search with only longest prefix
    """

    def __init__(
        self,
        decoder,
        joint,
        blank_index: int,
        beam_size: int,
        maes_num_steps: int,
        maes_expansion_beta: int,
        maes_expansion_gamma: int,
        preserve_alignments=False,
        ngram_lm_model: Optional[str | Path] = None,
        ngram_lm_alpha: float = 0.0,
        blank_lm_score_mode: Optional[str | BlankLMScoreMode] = BlankLMScoreMode.NO_SCORE,
        pruning_mode: Optional[str | PruningMode] = PruningMode.EARLY,
        allow_cuda_graphs: Optional[bool] = True,
    ):
        """
        Init method.
        Args:
            decoder: Prediction network from RNN-T
            joint: Joint module from RNN-T
            blank_index: index of blank symbol
            maes_num_steps:  Number of adaptive steps to take. From the paper, 2 steps is generally sufficient. int > 1.
            maes_expansion_beta: Maximum number of prefix expansions allowed, in addition to the beam size.
                Effectively, the number of hypothesis = beam_size + maes_expansion_beta. Must be an int >= 0,
                and affects the speed of inference since large values will perform large beam search in the next step.
            maes_expansion_gamma: Float pruning threshold used in the prune-by-value step when computing the expansions.
                The default (2.3) is selected from the paper. It performs a comparison
                (max_log_prob - gamma <= log_prob[v]) where v is all vocabulary indices in the Vocab set and max_log_prob
                is the "most" likely token to be predicted. Gamma therefore provides a margin of additional tokens which
                can be potential candidates for expansion apart from the "most likely" candidate.
                Lower values will reduce the number of expansions (by increasing pruning-by-value, thereby improving speed
                but hurting accuracy). Higher values will increase the number of expansions (by reducing pruning-by-value,
                thereby reducing speed but potentially improving accuracy). This is a hyper parameter to be experimentally
                tuned on a validation set.
            preserve_alignments: if alignments are needed
            ngram_lm_model: path to the NGPU-LM n-gram LM model: .arpa or .nemo formats
            ngram_lm_alpha: weight for the n-gram LM scores
            blank_lm_score_mode: mode for scoring blank symbol with LM
            pruning_mode: mode for pruning hypotheses with LM
            allow_cuda_graphs: whether to allow CUDA graphs
        """

        super().__init__()
        self.decoder = decoder
        self.joint = joint
        self._blank_index = blank_index

        self.beam_size = beam_size
        self.maes_num_steps = maes_num_steps
        self.maes_expansion_beta = maes_expansion_beta
        self.maes_expansion_gamma = maes_expansion_gamma
        self.preserve_alignments = preserve_alignments
        self._SOS = self._blank_index
        self.pruning_mode = pruning_mode
        self.blank_lm_score_mode = blank_lm_score_mode

        self.maes_num_expansions = self.beam_size + self.maes_expansion_beta

        if self.preserve_alignments:
            raise NotImplementedError("Preserve alignments is not supported")

        if allow_cuda_graphs:
            logging.info("CUDA Graphs are unsupported for `maes_batch`; preceeding pure pytorch decoding")

        if ngram_lm_model is not None:
            expected_blank_index = self.joint.num_classes_with_blank - self.joint.num_extra_outputs - 1
            if self._blank_index != expected_blank_index:
                raise ValueError(f"Invalid blank index: expected {expected_blank_index}, got {self._blank_index}")
            self.ngram_lm_batch = ngram_lm_model

            self.pruning_mode = PruningMode.EARLY if pruning_mode is None else PruningMode(pruning_mode)
            self.blank_lm_score_mode = (
                BlankLMScoreMode.LM_WEIGHTED_FULL
                if blank_lm_score_mode is None
                else BlankLMScoreMode(blank_lm_score_mode)
            )
        else:
            self.ngram_lm_batch = None
            self.blank_lm_score_mode = None
        self.ngram_lm_alpha = ngram_lm_alpha

    def batched_modified_adaptive_expansion_search_torch(
        self,
        encoder_output: torch.Tensor,
        encoder_output_length: torch.Tensor,
    ) -> BatchedBeamHyps:
        """
        Pure PyTorch implementation

        Args:
            encoder_output: output from the encoder
            encoder_output_length: lengths of the utterances in `encoder_output`
        """
        batch_size, max_time, _unused = encoder_output.shape
        device = encoder_output.device

        encoder_output_projected = self.joint.project_encoder(encoder_output)
        float_dtype = encoder_output_projected.dtype

        # init empty batched hypotheses
        batched_hyps = BatchedBeamHyps(
            batch_size=batch_size,
            beam_size=self.beam_size,
            blank_index=self._blank_index,
            init_length=max_time * (self.maes_num_steps + 1) if self.maes_num_steps is not None else max_time,
            device=device,
            float_dtype=float_dtype,
            store_prefix_hashes=True,
        )

        last_labels_wb = torch.full(
            [batch_size, self.beam_size], fill_value=self._SOS, device=device, dtype=torch.long
        )

        batch_indices = (
            torch.arange(batch_size, device=device)[:, None].expand(batch_size, self.beam_size).clone()
        )  # size: batch_size x beam_size
        beam_indices = (
            torch.arange(self.beam_size, device=device)[None, :].expand(batch_size, self.beam_size).clone()
        )  # size: batch_size x beam_size
        expansion_beam_indices = (
            torch.arange(self.beam_size, device=device)[None, :, None]
            .expand(batch_size, self.beam_size, self.maes_num_expansions)
            .clone()
        )  # size: batch_size x beam_size x beam_size + maes_expansion_beta

        time_indices = torch.zeros_like(batch_indices)
        safe_time_indices = torch.zeros_like(time_indices)
        last_timesteps = (encoder_output_length - 1)[:, None].expand(batch_size, self.beam_size)
        active_mask = time_indices <= last_timesteps

        # setup N-gram LM if available
        if self.ngram_lm_batch is not None:
            self.ngram_lm_batch.to(device)
            batch_lm_states = self.ngram_lm_batch.get_init_states(batch_size=batch_size * self.beam_size, bos=True)
            lm_scores, batch_lm_states_candidates = self.ngram_lm_batch.advance(
                states=batch_lm_states
            )  # vocab_size_no_blank
            lm_scores = lm_scores.to(dtype=float_dtype).view(batch_size, self.beam_size, -1) * self.ngram_lm_alpha

        decoder_output, decoder_state, *_ = self.decoder.predict(
            last_labels_wb.view(-1, 1), None, add_sos=False, batch_size=batch_size * self.beam_size
        )
        # do not recalculate joint projection
        decoder_output = self.joint.project_prednet(decoder_output)

        while active_mask.any():  # frames loop
            to_update = active_mask.clone()  # mask for expansions loop

            # step 1: get joint output
            logits = (
                self.joint.joint_after_projection(
                    encoder_output_projected[batch_indices.flatten(), safe_time_indices.flatten()].unsqueeze(1),
                    decoder_output,
                )
                .squeeze(1)
                .squeeze(1)
            )
            logps = torch.log_softmax(logits, dim=-1).view(batch_size, self.beam_size, -1)

            # step 2: perform prefix search
            updated_logps = self.combine_scores(logps, lm_scores) if self.ngram_lm_batch is not None else logps
            batched_hyps.recombine_prefixes(updated_logps, active_mask)

            expansion_steps = 0
            # step 3: performs `maes_num_steps` non-blank expansions
            while to_update.any() and expansion_steps < self.maes_num_steps:  # expansions loop
                # step 3.1: get `maes_num_expansion` best expansions (in total beam x maes_num_expansion expansions)
                if self.ngram_lm_batch is None:
                    # step 3.1.1: choose topk expansions (beam x beam hypotheses for each sample)
                    label_logps, next_labels = logps.topk(self.maes_num_expansions, dim=-1, largest=True, sorted=True)
                    next_hyps_probs = batched_hyps.scores.unsqueeze(-1) + label_logps

                    # step 3.1.2: prune with threshold parameter gamma
                    next_hyps_probs[
                        next_hyps_probs <= next_hyps_probs.max(dim=-1, keepdim=True).values - self.maes_expansion_gamma
                    ] = INACTIVE_SCORE
                else:
                    next_labels, next_hyps_probs = self.topk_lm(batched_hyps, lm_scores, logps)

                # step 3.2: get `beam` best expansions
                # step 3.2.1: mask inactive hypotheses
                next_labels = torch.where(to_update.unsqueeze(-1), next_labels, NON_EXISTENT_LABEL_VALUE)
                next_hyps_probs = torch.where(to_update.unsqueeze(-1), next_hyps_probs, INACTIVE_SCORE)

                # step 3.2.2: remove duplicate hypotheses
                next_hyps_probs = batched_hyps.remove_duplicates(next_labels, next_hyps_probs)

                # step 3.2.3: add hypotheses from the previous expansion steps of current frame hypotheses to the beam.
                # Expansions from step s are compared against the top beam expansions from steps 1 to s-1.
                next_hyps_probs[..., -1] = torch.where(to_update, next_hyps_probs[..., -1], batched_hyps.scores)

                # step 3.2.4: get top-k expansions
                next_hyps_probs, idx = next_hyps_probs.view(batch_size, -1).topk(
                    self.beam_size, dim=-1, largest=True, sorted=True
                )
                next_labels = next_labels.view(batch_size, -1)[batch_indices, idx]
                hyp_indices = expansion_beam_indices.view(batch_size, -1)[batch_indices, idx]

                # step 3.3: update batched beam hypotheses structure
                batched_hyps.add_results_(hyp_indices, next_labels, next_hyps_probs)

                # step 3.4: update
                last_labels_wb = torch.where(next_labels >= 0, next_labels, self._blank_index)
                preserve_state = last_labels_wb == self._blank_index

                # size: decoder_output [(B x Beam), 1, Dim]
                # size: state tuple, each is of [Layers, (BxBeam), Dim]
                # step 3.5: update decoder + lm state
                # step 3.5.1: storing current decoder output and states of extended hypotheses
                prev_decoder_output = torch.gather(
                    decoder_output.view(batch_size, self.beam_size, 1, -1),
                    dim=1,
                    index=hyp_indices[:, :, None, None].expand(
                        batch_size, self.beam_size, 1, decoder_output.shape[-1]
                    ),
                ).view(batch_size * self.beam_size, 1, -1)
                prev_decoder_state = self.decoder.batch_aggregate_states_beam(
                    decoder_state, batch_size, self.beam_size, hyp_indices
                )

                # step 3.5.2: get next decoder output and states for extended hypotheses
                decoder_output, decoder_state, *_ = self.decoder.predict(
                    last_labels_wb.view(-1, 1),
                    prev_decoder_state,
                    add_sos=False,
                    batch_size=batch_size * self.beam_size,
                )
                decoder_output = self.joint.project_prednet(decoder_output)  # do not recalculate joint projection

                # step 3.5.3: update decoder state and output only for non-blank and active hypotheses
                decoder_output = torch.where(
                    preserve_state.view(-1)[:, None, None], prev_decoder_output, decoder_output
                )
                self.decoder.batch_replace_states_mask(
                    src_states=prev_decoder_state, dst_states=decoder_state, mask=preserve_state.view(-1)
                )

                if self.ngram_lm_batch is not None:
                    # batch_lm_states: size: [(batch_size x beam_size)]
                    # batch_lm_states_candidates: [(batch_size x beam_size) x V (without blank)]
                    batch_lm_states_candidates = torch.gather(
                        batch_lm_states_candidates.view(batch_size, self.beam_size, -1),
                        dim=1,
                        index=hyp_indices[:, :, None].expand(
                            batch_size, self.beam_size, batch_lm_states_candidates.shape[-1]
                        ),
                    )
                    batch_lm_states_prev = torch.gather(
                        batch_lm_states.view(batch_size, self.beam_size), dim=1, index=hyp_indices
                    )
                    last_labels_wb_blank_replaced = torch.where(preserve_state, 0, last_labels_wb)

                    batch_lm_states = torch.gather(
                        batch_lm_states_candidates, dim=-1, index=last_labels_wb_blank_replaced.unsqueeze(-1)
                    ).squeeze(-1)
                    batch_lm_states = torch.where(preserve_state, batch_lm_states_prev, batch_lm_states).view(-1)

                    lm_scores, batch_lm_states_candidates = self.ngram_lm_batch.advance(
                        states=batch_lm_states
                    )  # vocab_size_no_blank
                    lm_scores = (
                        lm_scores.to(dtype=float_dtype).view(batch_size, self.beam_size, -1) * self.ngram_lm_alpha
                    )

                # step 3.6: get log-probs for next expansion step
                logits = self.joint.joint_after_projection(
                    encoder_output_projected[batch_indices.flatten(), safe_time_indices.flatten()].unsqueeze(1),
                    decoder_output,
                )
                logps = torch.log_softmax(logits, dim=-1).squeeze(1).squeeze(1).view(batch_size, self.beam_size, -1)
                to_update = torch.logical_and(to_update, last_labels_wb != self._blank_index)

                expansion_steps += 1
            if to_update.any():
                # step 4: force blank to active hypotheses
                next_hyps_probs = torch.where(to_update, batched_hyps.scores + logps[..., -1], batched_hyps.scores)
                next_labels = torch.where(to_update, self._blank_index, -1)
                batched_hyps.add_results_(beam_indices, next_labels, next_hyps_probs)

            # step 5: update time indices + active mask
            time_indices += 1
            active_mask = time_indices <= last_timesteps
            safe_time_indices = torch.where(active_mask, time_indices, last_timesteps)

        return batched_hyps

    def combine_scores(self, log_probs, lm_scores):
        """
        Combines acoustic model log probabilities with language model scores based on the specified blank LM score mode.

        Args:
            log_probs (torch.Tensor): Log probabilities from the acoustic model.
                Shape: (..., vocab_size), where the last dimension corresponds to the vocabulary size.
            lm_scores (torch.Tensor): Scores from the language model.
                Shape: (..., vocab_size - 1), excluding the blank token.

        Returns:
            torch.Tensor: Combined scores with the same shape as `log_probs`.

        Raises:
            NotImplementedError: If the `blank_lm_score_mode` is not supported.
        """
        res = log_probs.clone()
        if self.blank_lm_score_mode is BlankLMScoreMode.NO_SCORE:
            # choosing topk from acoustic and Ngram models
            res[..., :-1] += lm_scores
        else:
            blank_logprob = log_probs[..., -1]
            non_blank_logprob = torch.log1p(-torch.clamp(torch.exp(blank_logprob), max=1.0 - 1e-6))
            res[..., :-1] += non_blank_logprob.unsqueeze(-1) * self.ngram_lm_alpha + lm_scores
            res[..., -1] *= 1 + self.ngram_lm_alpha

        return res

    def topk_lm(self, batched_hyps, lm_scores, log_probs):
        """
        Performs top-k selection and pruning for language model (LM) and automatic speech recognition (ASR) outputs
        based on the specified pruning and blank scoring modes.
        Args:
            batched_hyps (object): Hypotheses from the ASR model, containing scores and other relevant information.
            lm_scores (Tensor): Precomputed language model scores for the current batch.
            log_probs (Tensor): Log probabilities from the ASR model.
        Returns:
            Tuple[Tensor, Tensor]:
                - labels (Tensor): The top-k labels selected after pruning and scoring.
                - total_logps (Tensor): The corresponding total log probabilities for the selected labels.
        Raises:
            NotImplementedError: If the combination of `blank_lm_score_mode` and `pruning_mode` is not implemented.
        """

        match self.pruning_mode, self.blank_lm_score_mode:
            case PruningMode.LATE, BlankLMScoreMode.NO_SCORE | BlankLMScoreMode.LM_WEIGHTED_FULL:
                # step 1: combining LM and ASR outputs + choosing top `beam` most probable
                log_probs = self.combine_scores(log_probs, lm_scores)
                label_logps, labels = log_probs.topk(
                    self.beam_size + self.maes_expansion_beta, dim=-1, largest=True, sorted=True
                )

                # step 2: pruning with threshold gamma
                total_logps = batched_hyps.scores.unsqueeze(-1) + label_logps
                total_logps[
                    total_logps <= total_logps.max(dim=-1, keepdim=True).values - self.maes_expansion_gamma
                ] = INACTIVE_SCORE

            case PruningMode.EARLY, BlankLMScoreMode.NO_SCORE:
                # step 1: choosing topk from ASR output
                label_logps, labels = log_probs.topk(
                    self.beam_size + self.maes_expansion_beta, dim=-1, largest=True, sorted=True
                )

                # step 2: pruning with threshold gamma
                total_logps = batched_hyps.scores.unsqueeze(-1) + label_logps
                total_logps[
                    total_logps <= total_logps.max(dim=-1, keepdim=True).values - self.maes_expansion_gamma
                ] = INACTIVE_SCORE

                # step 3: adding scores from ngram LM
                masked_labels = torch.where(labels == self._blank_index, 0, labels)
                total_logps = torch.where(
                    labels == self._blank_index,
                    total_logps,
                    total_logps + torch.gather(lm_scores, dim=-1, index=masked_labels),
                )

            case PruningMode.EARLY, BlankLMScoreMode.LM_WEIGHTED_FULL:
                # step 1: choosing topk from ASR output
                label_logps, labels = log_probs.topk(
                    self.beam_size + self.maes_expansion_beta, dim=-1, largest=True, sorted=True
                )

                # step 2: pruning with threshold gamma
                total_logps = batched_hyps.scores.unsqueeze(-1) + label_logps
                label_logps[
                    total_logps <= total_logps.max(dim=-1, keepdim=True).values - self.maes_expansion_gamma
                ] = INACTIVE_SCORE

                # step 3: adding scores from ngram LM
                blank_logprob = log_probs[..., -1]
                non_blank_logprob = torch.log1p(-torch.clamp(torch.exp(blank_logprob), max=1.0 - 1e-6))

                masked_labels = torch.where(labels == self._blank_index, 0, labels)
                total_logps = torch.where(
                    labels == self._blank_index,
                    total_logps + label_logps * (1 + self.ngram_lm_alpha),
                    total_logps
                    + label_logps
                    + non_blank_logprob.unsqueeze(-1) * self.ngram_lm_alpha
                    + torch.gather(lm_scores, dim=-1, index=masked_labels),
                )

            case _:
                raise NotImplementedError(
                    f"Unsupported pruning mode {self.pruning_mode} or blank LM score mode {self.blank_lm_score_mode}"
                )

        return labels, total_logps

    def __call__(
        self,
        x: torch.Tensor,
        out_len: torch.Tensor,
    ) -> BatchedBeamHyps:
        return self.batched_modified_adaptive_expansion_search_torch(encoder_output=x, encoder_output_length=out_len)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# code taken from
# https://github.com/NVIDIA/Megatron-LM/blob/main/megatron/data/image_folder.py

import os
import os.path
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import numpy as np
from PIL import Image
from nemo.utils import logging

try:
    from torchvision.datasets import VisionDataset

    TORCHVISION_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    TORCHVISION_AVAILABLE = False


def has_file_allowed_extension(filename: str, extensions: Tuple[str, ...]) -> bool:
    """Checks if a file is an allowed extension.
    Args:
        filename (string): path to a file
        extensions (tuple of strings): extensions to consider (lowercase)
    Returns:
        bool: True if the filename ends with one of given extensions
    """
    return filename.lower().endswith(extensions)


def is_image_file(filename: str) -> bool:
    """Checks if a file is an allowed image extension.
    Args:
        filename (string): path to a file
    Returns:
        bool: True if the filename ends with a known image extension
    """
    return has_file_allowed_extension(filename, IMG_EXTENSIONS)


def make_dataset(
    directory: str,
    class_to_idx: Dict[str, int],
    data_per_class_fraction: float,
    extensions: Optional[Tuple[str, ...]] = None,
    is_valid_file: Optional[Callable[[str], bool]] = None,
) -> List[Tuple[str, int]]:
    """Generates a list of samples of a form (path_to_sample, class).
    Args:
        directory (str): root dataset directory
        class_to_idx (Dict[str, int]): dictionary mapping class name to class index
        extensions (optional): A list of allowed extensions.
            Either extensions or is_valid_file should be passed. Defaults to None.
        is_valid_file (optional): A function that takes path of a file
            and checks if the file is a valid file
            (used to check of corrupt files) both extensions and
            is_valid_file should not be passed. Defaults to None.
    Raises:
        ValueError: In case ``extensions`` and ``is_valid_file`` are None or both are not None.
    Returns:
        List[Tuple[str, int]]: samples of a form (path_to_sample, class)
    """
    instances = []
    directory = os.path.expanduser(directory)
    both_none = extensions is None and is_valid_file is None
    both_something = extensions is not None and is_valid_file is not None
    if both_none or both_something:
        raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
    if extensions is not None:

        def is_valid_file(x: str) -> bool:
            return has_file_allowed_extension(x, cast(Tuple[str, ...], extensions))

    is_valid_file = cast(Callable[[str], bool], is_valid_file)
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        local_instances = []
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                if is_valid_file(path):
                    item = path, class_index
                    local_instances.append(item)

        instances.extend(local_instances[0 : int(len(local_instances) * data_per_class_fraction)])

    return instances


if TORCHVISION_AVAILABLE:

    class DatasetFolder(VisionDataset):
        """A generic data loader where the samples are arranged in this way: ::
            root/class_x/xxx.ext
            root/class_x/xxy.ext
            root/class_x/[...]/xxz.ext
            root/class_y/123.ext
            root/class_y/nsdf3.ext
            root/class_y/[...]/asd932_.ext
        Args:
            root (string): Root directory path.
            loader (callable): A function to load a sample given its path.
            extensions (tuple[string]): A list of allowed extensions.
                both extensions and is_valid_file should not be passed.
            transform (callable, optional): A function/transform that takes in
                a sample and returns a transformed version.
                E.g, ``transforms.RandomCrop`` for images.
            target_transform (callable, optional): A function/transform that takes
                in the target and transforms it.
            is_valid_file (callable, optional): A function that takes path of a file
                and check if the file is a valid file (used to check of corrupt files)
                both extensions and is_valid_file should not be passed.
         Attributes:
            classes (list): List of the class names sorted alphabetically.
            class_to_idx (dict): Dict with items (class_name, class_index).
            samples (list): List of (sample path, class_index) tuples
            targets (list): The class_index value for each image in the dataset
        """

        def __init__(
            self,
            root: str,
            loader: Callable[[str], Any],
            extensions: Optional[Tuple[str, ...]] = None,
            transform: Optional[Callable] = None,
            target_transform: Optional[Callable] = None,
            classes_fraction=1.0,
            data_per_class_fraction=1.0,
            is_valid_file: Optional[Callable[[str], bool]] = None,
        ) -> None:
            super(DatasetFolder, self).__init__(root, transform=transform, target_transform=target_transform)
            self.classes_fraction = classes_fraction
            self.data_per_class_fraction = data_per_class_fraction
            classes, class_to_idx = self._find_classes(self.root)
            samples = self.make_dataset(
                self.root, class_to_idx, self.data_per_class_fraction, extensions, is_valid_file
            )
            if len(samples) == 0:
                msg = "Found 0 files in subfolders of: {}\n".format(self.root)
                if extensions is not None:
                    msg += "Supported extensions are: {}".format(",".join(extensions))
                raise RuntimeError(msg)

            self.loader = loader
            self.extensions = extensions
            self.total = len(samples)
            self.classes = classes
            self.class_to_idx = class_to_idx
            self.samples = samples
            self.targets = [s[1] for s in samples]

        @staticmethod
        def make_dataset(
            directory: str,
            class_to_idx: Dict[str, int],
            data_per_class_fraction: float,
            extensions: Optional[Tuple[str, ...]] = None,
            is_valid_file: Optional[Callable[[str], bool]] = None,
        ) -> List[Tuple[str, int]]:
            return make_dataset(
                directory, class_to_idx, data_per_class_fraction, extensions=extensions, is_valid_file=is_valid_file
            )

        def _find_classes(self, dir: str) -> Tuple[List[str], Dict[str, int]]:
            """
            Finds the class folders in a dataset.
            Args:
                dir (string): Root directory path.
            Returns:
                tuple: (classes, class_to_idx) where classes are relative to (dir), and class_to_idx is a dictionary.
            Ensures:
                No class is a subdirectory of another.
            """
            all_classes = [d.name for d in os.scandir(dir) if d.is_dir()]
            classes = all_classes[0 : int(len(all_classes) * self.classes_fraction)]
            classes.sort()
            class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
            return classes, class_to_idx

        def __getitem__(self, index: int) -> Tuple[Any, Any]:
            """
            Args:
                index (int): Index
            Returns:
                tuple: (sample, target) where target is class_index of the target class.
            """
            curr_index = index
            for x in range(self.total):
                try:
                    path, target = self.samples[curr_index]
                    sample = self.loader(path)
                    break
                except Exception as e:
                    curr_index = np.random.randint(0, self.total)

            if self.transform is not None:
                sample = self.transform(sample)
            if self.target_transform is not None:
                target = self.target_transform(target)

            return sample, target

        def __len__(self) -> int:
            return len(self.samples)


else:

    class DatasetFolder:
        def __init__(self):
            super().__init__()
            logging.error("Torchvision not found but required.")


IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif', '.tiff', '.webp')


def pil_loader(path: str) -> Image.Image:
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, 'rb') as f:
        img = Image.open(f)
        return img.convert('RGB')


# TODO: specify the return type
def accimage_loader(path: str) -> Any:
    import accimage

    try:
        return accimage.Image(path)
    except IOError:
        # Potentially a decoding problem, fall back to PIL.Image
        return pil_loader(path)


def default_loader(path: str) -> Any:
    from torchvision import get_image_backend

    if get_image_backend() == 'accimage':
        return accimage_loader(path)
    else:
        return pil_loader(path)


class ImageFolder(DatasetFolder):
    """A generic data loader where the images are arranged in this way: ::
        root/dog/xxx.png
        root/dog/xxy.png
        root/dog/[...]/xxz.png
        root/cat/123.png
        root/cat/nsdf3.png
        root/cat/[...]/asd932_.png
    Args:
        root (string): Root directory path.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        loader (callable, optional): A function to load an image given its path.
        is_valid_file (callable, optional): A function that takes path of an Image file
            and check if the file is a valid file (used to check of corrupt files)
     Attributes:
        classes (list): List of the class names sorted alphabetically.
        class_to_idx (dict): Dict with items (class_name, class_index).
        imgs (list): List of (image path, class_index) tuples
    """

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        classes_fraction=1.0,
        data_per_class_fraction=1.0,
        loader: Callable[[str], Any] = default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ):
        super(ImageFolder, self).__init__(
            root,
            loader,
            IMG_EXTENSIONS if is_valid_file is None else None,
            transform=transform,
            target_transform=target_transform,
            classes_fraction=classes_fraction,
            data_per_class_fraction=data_per_class_fraction,
            is_valid_file=is_valid_file,
        )
        self.imgs = self.samples



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lightning.pytorch as pl
import torch
from lightning.pytorch.callbacks.callback import Callback
from megatron.core.transformer import TransformerConfig


class DeepEPCallback(Callback):
    """
    A PyTorch Lightning callback to enable DeepEP if the hardware is supported.
    Per official documentation https://github.com/deepseek-ai/DeepEP,
    DeepEP is supported for Ampere (SM80) and Hopper (SM90) GPUs.

    Adding this callback is equivalent to setting the following flags in the recipe function:

    recipe.model.config.moe_token_dispatcher_type = "flex"
    recipe.model.config.moe_enable_deepep = True
    recipe.model.config.moe_shared_expert_overlap = False

    Since the recipe function may be run on a different machine, this callback is needed so that
    configs are set during run time.
    """

    def setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule, stage: str) -> None:
        """Enable DeepEP if GPU is Ampere or Hopper"""
        if torch.cuda.get_device_properties(0).major not in [8, 9]:
            return

        if hasattr(trainer.model, "config") and isinstance(trainer.model.config, TransformerConfig):
            self._apply_deepep_cfgs(trainer.model.config)
            if hasattr(trainer.model, '__io__'):
                self._apply_deepep_cfgs(trainer.model.__io__.config)

    def _apply_deepep_cfgs(self, dest_cfg):
        # apply optimizations into dest_cfg
        dest_cfg.moe_token_dispatcher_type = "flex"
        dest_cfg.moe_enable_deepep = True
        dest_cfg.moe_shared_expert_overlap = False



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.collections.tts.modules import submodules


@pytest.mark.unit
def test_conditional_layer_norm():

    # NLP Example
    batch, sentence_length, embedding_dim = 20, 5, 10
    embedding = torch.randn(batch, sentence_length, embedding_dim)
    ln = torch.nn.LayerNorm(embedding_dim)
    cln = submodules.ConditionalLayerNorm(embedding_dim)
    assert torch.all(ln(embedding) == cln(embedding))

    weight = torch.nn.Parameter(torch.randn(embedding_dim))
    bias = torch.nn.Parameter(torch.randn(embedding_dim))
    ln.weight, ln.bias = weight, bias
    cln.weight, cln.bias = weight, bias
    assert torch.all(ln(embedding) == cln(embedding))  # Simulate trained weights

    # Image Example
    N, C, H, W = 20, 5, 10, 10
    image = torch.randn(N, C, H, W)
    ln = torch.nn.LayerNorm([C, H, W])
    cln = submodules.ConditionalLayerNorm([C, H, W])
    assert torch.all(ln(image) == cln(image))

    weight = torch.nn.Parameter(torch.randn(C, H, W))
    bias = torch.nn.Parameter(torch.randn(C, H, W))
    ln.weight, ln.bias = weight, bias
    cln.weight, cln.bias = weight, bias
    assert torch.all(ln(image) == cln(image))  # Simulate trained weights



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright    2023  Xiaomi Corp.        (authors: Wei Kang)
#
# See ../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The script was obtained and modified from Icefall repo:
# https://github.com/k2-fsa/icefall/blob/aac7df064a6d1529f3bf4acccc6c550bd260b7b3/icefall/context_graph.py


import os
import shutil
from collections import deque
from typing import Dict, List, Optional
import numpy as np


class ContextState:
    """The state in ContextGraph"""

    def __init__(
        self,
        id: int,
        token: int,
        token_score: float,
        node_score: float,
        output_score: float,
        is_end: bool,
        level: int,
        phrase: str = "",
        ac_threshold: float = 1.0,
    ):
        """Create a ContextState.

        Args:
          id:
            The node id, only for visualization now. A node is in [0, graph.num_nodes).
            The id of the root node is always 0.
          token:
            The token id.
          token_score:
            The bonus for each token during decoding, which will hopefully
            boost the token up to survive beam search.
          node_score:
            The accumulated bonus from root of graph to current node, it will be
            used to calculate the score for fail arc.
          output_score:
            The total scores of matched phrases, sum of the node_score of all
            the output node for current node.
          is_end:
            True if current token is the end of a context.
          level:
            The distance from current node to root.
          phrase:
            The context phrase of current state, the value is valid only when
            current state is end state (is_end == True).
          ac_threshold:
            The acoustic threshold (probability) of current context phrase, the
            value is valid only when current state is end state (is_end == True).
            Note: ac_threshold only used in keywords spotting.
        """
        self.id = id
        self.token = token
        self.token_score = token_score
        self.node_score = node_score
        self.output_score = output_score
        self.is_end = is_end
        self.level = level
        self.next = {}
        self.phrase = phrase
        self.ac_threshold = ac_threshold
        self.fail = None
        self.output = None


class ContextGraph:
    """The ContextGraph is modified from Aho-Corasick which is mainly
    a Trie with a fail arc for each node.
    See https://en.wikipedia.org/wiki/Aho%E2%80%93Corasick_algorithm for more details
    of Aho-Corasick algorithm.

    A ContextGraph contains some words / phrases that we expect to boost their
    scores during decoding. If the substring of a decoded sequence matches the word / phrase
    in the ContextGraph, we will give the decoded sequence a bonus to make it survive
    beam search.
    """

    def __init__(self, context_score: float, depth_scaling: float = 1.0, ac_threshold: float = 1.0):
        """Initialize a ContextGraph with the given ``context_score``.

        A root node will be created (**NOTE:** the token of root is hardcoded to -1).

        Args:
          context_score:
            The bonus score for each token(note: NOT for each word/phrase, it means longer
            word/phrase will have larger bonus score, they have to be matched though).
            Note: This is just the default score for each token, the users can manually
            specify the context_score for each word/phrase (i.e. different phrase might
            have different token score).
          depth_scaling:
            The depth scaling factor for each token [1, inf), it is used to give a larger score for all tokens after the first one.
          ac_threshold:
            The acoustic threshold (probability) to trigger the word/phrase, this argument
            is used only when applying the graph to keywords spotting system.
        """
        self.context_score = context_score
        self.depth_scaling = depth_scaling
        self.ac_threshold = ac_threshold
        self.num_nodes = 0
        self.root = ContextState(
            id=self.num_nodes,
            token=-1,
            token_score=0,
            node_score=0,
            output_score=0,
            is_end=False,
            level=0,
        )
        self.root.fail = self.root

    def _fill_fail_output(self):
        """This function fills the fail arc for each trie node, it can be computed
        in linear time by performing a breadth-first search starting from the root.
        See https://en.wikipedia.org/wiki/Aho%E2%80%93Corasick_algorithm for the
        details of the algorithm.
        """
        queue = deque()
        for token, node in self.root.next.items():
            node.fail = self.root
            queue.append(node)
        while queue:
            current_node = queue.popleft()
            for token, node in current_node.next.items():
                fail = current_node.fail
                if token in fail.next:
                    fail = fail.next[token]
                else:
                    fail = fail.fail
                    while token not in fail.next:
                        fail = fail.fail
                        if fail.token == -1:  # root
                            break
                    if token in fail.next:
                        fail = fail.next[token]
                node.fail = fail
                # fill the output arc
                output = node.fail
                while not output.is_end:
                    output = output.fail
                    if output.token == -1:  # root
                        output = None
                        break
                node.output = output
                node.output_score += 0 if output is None else output.output_score
                queue.append(node)

    def build(
        self,
        token_ids: List[List[int]],
        phrases: Optional[List[str]] = None,
        scores: Optional[List[float]] = None,
        ac_thresholds: Optional[List[float]] = None,
        uniform_weights: Optional[bool] = False,
    ):
        """Build the ContextGraph from a list of token list.
        It first build a trie from the given token lists, then fill the fail arc
        for each trie node.

        See https://en.wikipedia.org/wiki/Trie for how to build a trie.

        Args:
          token_ids:
            The given token lists to build the ContextGraph, it is a list of
            token list, the token list contains the token ids
            for a word/phrase. The token id could be an id of a char
            (modeling with single Chinese char) or an id of a BPE
            (modeling with BPEs).
          phrases:
            The given phrases, they are the original text of the token_ids, the
            length of `phrases` MUST be equal to the length of `token_ids`.
          scores:
            The customize boosting score(token level) for each word/phrase,
            0 means using the default value (i.e. self.context_score).
            It is a list of floats, and the length of `scores` MUST be equal to
            the length of `token_ids`.
          ac_thresholds:
            The customize trigger acoustic threshold (probability) for each phrase,
            0 means using the default value (i.e. self.ac_threshold). It is
            used only when this graph applied for the keywords spotting system.
            The length of `ac_threshold` MUST be equal to the length of `token_ids`.
          uniform_weights:
            If True, the weights will be distributed uniformly for all tokens as in Icefall.

        Note: The phrases would have shared states, the score of the shared states is
              the MAXIMUM value among all the tokens sharing this state.
        """
        num_phrases = len(token_ids)
        if phrases is not None:
            assert len(phrases) == num_phrases, (len(phrases), num_phrases)
        if scores is not None:
            assert len(scores) == num_phrases, (len(scores), num_phrases)
        if ac_thresholds is not None:
            assert len(ac_thresholds) == num_phrases, (len(ac_thresholds), num_phrases)

        for index, tokens in enumerate(token_ids):
            phrase = "" if phrases is None else phrases[index]
            score = 0.0 if scores is None else scores[index]
            ac_threshold = 0.0 if ac_thresholds is None else ac_thresholds[index]
            node = self.root
            # If has customized score using the customized token score, otherwise
            # using the default score
            context_score = self.context_score if score == 0.0 else score
            threshold = self.ac_threshold if ac_threshold == 0.0 else ac_threshold
            for i, token in enumerate(tokens):
                if token not in node.next:
                    if i > 0 and not uniform_weights:
                        token_score = context_score * self.depth_scaling + np.log(
                            i + 1
                        )  # depth scaling is used to give a larger score for all tokens after the first one
                    else:
                        token_score = context_score
                    self.num_nodes += 1
                    is_end = i == len(tokens) - 1
                    node_score = node.node_score + token_score
                    node.next[token] = ContextState(
                        id=self.num_nodes,
                        token=token,
                        token_score=token_score,
                        node_score=node_score,
                        output_score=node_score if is_end else 0,
                        is_end=is_end,
                        level=i + 1,
                        phrase=phrase if is_end else "",
                        ac_threshold=threshold if is_end else 0.0,
                    )
                else:
                    # node exists, get the score of shared state.
                    token_score = max(context_score, node.next[token].token_score)
                    node.next[token].token_score = token_score
                    node_score = node.node_score + token_score
                    node.next[token].node_score = node_score
                    is_end = i == len(tokens) - 1 or node.next[token].is_end
                    node.next[token].output_score = node_score if is_end else 0
                    node.next[token].is_end = is_end
                    if i == len(tokens) - 1:
                        node.next[token].phrase = phrase
                        node.next[token].ac_threshold = threshold
                node = node.next[token]
        self._fill_fail_output()

    def draw(
        self,
        title: Optional[str] = None,
        filename: Optional[str] = "",
        symbol_table: Optional[Dict[int, str]] = None,
    ) -> "Digraph":  # noqa
        """Visualize a ContextGraph via graphviz.

        Render ContextGraph as an image via graphviz, and return the Digraph object;
        and optionally save to file `filename`.
        `filename` must have a suffix that graphviz understands, such as
        `pdf`, `svg` or `png`.

        Note:
          You need to install graphviz to use this function::

            pip install graphviz

        Args:
           title:
              Title to be displayed in image, e.g. 'A simple FSA example'
           filename:
              Filename to (optionally) save to, e.g. 'foo.png', 'foo.svg',
              'foo.png'  (must have a suffix that graphviz understands).
           symbol_table:
              Map the token ids to symbols.
        Returns:
          A Diagraph from grahpviz.
        """

        try:
            import graphviz
        except Exception:
            print("You cannot use `to_dot` unless the graphviz package is installed.")
            raise

        graph_attr = {
            "rankdir": "LR",
            "size": "8.5,11",
            "center": "1",
            "orientation": "Portrait",
            "ranksep": "0.4",
            "nodesep": "0.25",
        }
        if title is not None:
            graph_attr["label"] = title

        default_node_attr = {
            "shape": "circle",
            "style": "bold",
            "fontsize": "14",
        }

        final_state_attr = {
            "shape": "doublecircle",
            "style": "bold",
            "fontsize": "14",
        }

        dot = graphviz.Digraph(name="Context Graph", graph_attr=graph_attr)

        seen = set()
        queue = deque()
        queue.append(self.root)
        # root id is always 0
        dot.node("0", label="0", **default_node_attr)
        dot.edge("0", "0", color="red")
        seen.add(0)

        while len(queue):
            current_node = queue.popleft()
            for token, node in current_node.next.items():
                if node.id not in seen:
                    node_score = f"{node.node_score:.2f}".rstrip("0").rstrip(".")
                    output_score = f"{node.output_score:.2f}".rstrip("0").rstrip(".")
                    label = f"{node.id}/({node_score}, {output_score})"
                    if node.is_end:
                        dot.node(str(node.id), label=label, **final_state_attr)
                    else:
                        dot.node(str(node.id), label=label, **default_node_attr)
                    seen.add(node.id)
                weight = f"{node.token_score:.2f}".rstrip("0").rstrip(".")
                label = str(token) if symbol_table is None else symbol_table[token]
                dot.edge(str(current_node.id), str(node.id), label=f"{label}/{weight}")
                dot.edge(
                    str(node.id),
                    str(node.fail.id),
                    color="red",
                )
                if node.output is not None:
                    dot.edge(
                        str(node.id),
                        str(node.output.id),
                        color="green",
                    )
                queue.append(node)

        if filename:
            _, extension = os.path.splitext(filename)
            if extension == "" or extension[0] != ".":
                raise ValueError("Filename needs to have a suffix like .png, .pdf, .svg: {}".format(filename))

            import tempfile

            with tempfile.TemporaryDirectory() as tmp_dir:
                temp_fn = dot.render(
                    filename="temp",
                    directory=tmp_dir,
                    format=extension[1:],
                    cleanup=True,
                )

                shutil.move(temp_fn, filename)

        return dot



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
from torchmetrics import ScaleInvariantSignalDistortionRatio

from nemo.collections.common.parts.utils import mask_sequence_tensor
from nemo.collections.tts.losses.audio_codec_loss import MaskedMAELoss, MaskedMSELoss, SISDRLoss


class TestAudioCodecLoss:
    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_masked_loss_l1(self):
        loss_fn = MaskedMAELoss()
        target = torch.tensor([[[1.0], [2.0], [0.0]], [[3.0], [0.0], [0.0]]]).transpose(1, 2)
        predicted = torch.tensor([[[0.5], [1.0], [0.0]], [[4.5], [0.0], [0.0]]]).transpose(1, 2)
        target_len = torch.tensor([2, 1])

        loss = loss_fn(predicted=predicted, target=target, target_len=target_len)

        assert loss == 1.125

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_masked_loss_l2(self):
        loss_fn = MaskedMSELoss()
        target = torch.tensor([[[1.0], [2.0], [4.0]], [[3.0], [0.0], [0.0]]]).transpose(1, 2)
        predicted = torch.tensor([[[0.5], [1.0], [4.0]], [[4.5], [0.0], [0.0]]]).transpose(1, 2)
        target_len = torch.tensor([3, 1])

        loss = loss_fn(predicted=predicted, target=target, target_len=target_len)

        assert loss == (4 / 3)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_si_sdr_loss(self):
        loss_fn = SISDRLoss()
        sdr_fn = ScaleInvariantSignalDistortionRatio(zero_mean=True)

        num_samples = 1000
        torch.manual_seed(100)
        target = torch.rand([1, num_samples])
        predicted = torch.rand([1, num_samples])
        target_len = torch.tensor([num_samples, num_samples])

        torch_si_sdr = sdr_fn(preds=predicted, target=target)
        loss = loss_fn(audio_real=target, audio_gen=predicted, audio_len=target_len)
        si_sdr = -loss

        torch.testing.assert_close(actual=si_sdr, expected=torch_si_sdr)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    def test_si_sdr_loss_batch(self):
        loss_fn = SISDRLoss()
        si_sdr_fn = ScaleInvariantSignalDistortionRatio(zero_mean=True)

        batch_size = 3
        num_samples = 1000
        torch.manual_seed(100)
        target = torch.rand([batch_size, num_samples])
        predicted = torch.rand([batch_size, num_samples])

        target_len = torch.tensor([500, 250, 900])
        target = mask_sequence_tensor(target, lengths=target_len)
        predicted = mask_sequence_tensor(predicted, lengths=target_len)

        torch_si_sdr = 0.0
        for i in range(batch_size):
            torch_si_sdr += si_sdr_fn(preds=predicted[i, : target_len[i]], target=target[i, : target_len[i]])
        torch_si_sdr /= batch_size

        loss = loss_fn(audio_real=target, audio_gen=predicted, audio_len=target_len)
        si_sdr = -loss

        torch.testing.assert_close(actual=si_sdr, expected=torch_si_sdr)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil
import tempfile

import lightning.pytorch as pl
import pytest
import torch
from omegaconf import DictConfig

from nemo.collections.asr.models import ASRModel, EncDecCTCModel


def getattr2(object, attr):
    if not '.' in attr:
        return getattr(object, attr)
    else:
        arr = attr.split('.')
        return getattr2(getattr(object, arr[0]), '.'.join(arr[1:]))


class TestASRLocalAttention:
    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_forward(self):
        asr_model = ASRModel.from_pretrained("stt_en_conformer_ctc_small")
        asr_model = asr_model.eval()

        len = 16000 * 60 * 30  # 30 minutes, OOM without local attention
        input_signal_long = torch.randn(size=(1, len), device=asr_model.device)
        length_long = torch.tensor([len], device=asr_model.device)

        # switch to local attn
        asr_model.change_attention_model(self_attention_model="rel_pos_local_attn", att_context_size=(64, 64))
        with torch.no_grad():
            asr_model.forward(input_signal=input_signal_long, input_signal_length=length_long)

        # switch context size only (keep local)
        asr_model.change_attention_model(att_context_size=(192, 192))
        with torch.no_grad():
            asr_model.forward(input_signal=input_signal_long, input_signal_length=length_long)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_change_save_restore(self):

        model = ASRModel.from_pretrained("stt_en_conformer_ctc_small")
        model.change_attention_model(self_attention_model="rel_pos_local_attn", att_context_size=(64, 64))
        attr_for_eq_check = ["encoder.self_attention_model", "encoder.att_context_size"]

        with tempfile.TemporaryDirectory() as restore_folder:
            with tempfile.TemporaryDirectory() as save_folder:
                save_folder_path = save_folder
                # Where model will be saved
                model_save_path = os.path.join(save_folder, f"{model.__class__.__name__}.nemo")
                model.save_to(save_path=model_save_path)
                # Where model will be restored from
                model_restore_path = os.path.join(restore_folder, f"{model.__class__.__name__}.nemo")
                shutil.copy(model_save_path, model_restore_path)
            # at this point save_folder should not exist
            assert save_folder_path is not None and not os.path.exists(save_folder_path)
            assert not os.path.exists(model_save_path)
            assert os.path.exists(model_restore_path)
            # attempt to restore
            model_copy = model.__class__.restore_from(
                restore_path=model_restore_path,
                map_location=None,
                strict=True,
                return_config=False,
                override_config_path=None,
            )

            assert model.num_weights == model_copy.num_weights
            if attr_for_eq_check is not None and len(attr_for_eq_check) > 0:
                for attr in attr_for_eq_check:
                    assert getattr2(model, attr) == getattr2(model_copy, attr)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "global_tokens",
        [0, 1, 4],
    )
    @pytest.mark.parametrize(
        "global_tokens_spacing",
        [1, 4],
    )
    def test_train(self, global_tokens, global_tokens_spacing):
        preprocessor_config = {'_target_': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor'}
        vocabulary = [
            ' ',
            'a',
            'b',
            'c',
            'd',
            'e',
            'f',
            'g',
            'h',
            'i',
            'j',
            'k',
            'l',
            'm',
            'n',
            'o',
            'p',
            'q',
            'r',
            's',
            't',
            'u',
            'v',
            'w',
            'x',
            'y',
            'z',
            "'",
        ]
        encoder_config = {
            '_target_': 'nemo.collections.asr.modules.ConformerEncoder',
            'feat_in': 64,
            'n_layers': 8,
            'd_model': 4,
            'self_attention_model': 'rel_pos_local_attn',
            'att_context_size': [128, 128],
            'global_tokens': global_tokens,
            'global_tokens_spacing': global_tokens_spacing,
        }
        decoder_config = {
            '_target_': 'nemo.collections.asr.modules.ConvASRDecoder',
            'feat_in': None,
            'num_classes': len(vocabulary),
            'vocabulary': vocabulary,
        }
        model_config = DictConfig(
            {
                'preprocessor': DictConfig(preprocessor_config),
                'encoder': DictConfig(encoder_config),
                'decoder': DictConfig(decoder_config),
                'optim': {'name': 'adamw'},
            }
        )

        class DummyDataset(torch.utils.data.Dataset):
            """Simply returns a single set of values."""

            def __init__(self, values):
                self.values = values

            def __len__(self):
                return 1

            def __getitem__(self, idx):
                return self.values

        input_signal = torch.randn(size=(1, 960000))
        input_length = torch.tensor([960000])
        target = torch.randint(size=(1, 280), low=0, high=28)
        target_length = torch.tensor([280])

        asr_model = EncDecCTCModel(cfg=model_config)
        asr_model.train()
        _ = asr_model.forward(input_signal=input_signal, input_signal_length=input_length)
        ## Explicitly pass acclerator as cpu, since deafult val in PTL >= 2.0 is auto and it picks cuda
        ## which further causes an error in all reduce at: https://github.com/NVIDIA/NeMo/blob/v1.18.1/nemo/collections/asr/modules/conformer_encoder.py#L462
        ## and in ConvASREncoder, SqueezeformerEncoder where device is CPU
        trainer = pl.Trainer(max_epochs=1, accelerator='cpu')
        trainer.fit(
            asr_model,
            train_dataloaders=torch.utils.data.DataLoader(
                DummyDataset([input_signal, input_length, target, target_length]),
                collate_fn=lambda x: x[0],
            ),
            val_dataloaders=torch.utils.data.DataLoader(
                DummyDataset([input_signal, input_length, target, target_length]),
                collate_fn=lambda x: x[0],
            ),
        )
        trainer.test(
            asr_model,
            dataloaders=torch.utils.data.DataLoader(
                DummyDataset([input_signal, input_length, target, target_length]),
                collate_fn=lambda x: x[0],
            ),
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyright: reportMissingImports=false
# pylint: disable=import-error,redefined-outer-name


import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Tuple

import pytest
import soundfile as sf
import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader, Dataset

from nemo.collections.asr.parts.mixins.diarization import DiarizeConfig, SpkDiarizationMixin


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torch.nn.Linear(1, 1)
        # Make the test deterministic and ensure positive outputs.
        with torch.no_grad():
            self.encoder.weight.fill_(1.0)
            self.encoder.bias.fill_(1.0)

        self.execution_count = 0
        self.flag_begin = False

    def forward(self, x):
        # Input: [1, 1] Output = [1, 1
        out = self.encoder(x)
        return out


class AudioPathDataset(Dataset):
    def __init__(self, audio_filepaths: List[str]):
        self._audio_filepaths = audio_filepaths

    def __len__(self) -> int:
        return len(self._audio_filepaths)

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, torch.Tensor]:
        samples, _ = sf.read(self._audio_filepaths[index], dtype="float32", always_2d=False)
        if hasattr(samples, "ndim") and samples.ndim == 2:
            samples = samples.mean(axis=1)
        waveform = torch.as_tensor(samples, dtype=torch.float32)
        length = torch.tensor(waveform.shape[0], dtype=torch.long)
        return waveform, length


def collate(batch):
    waveforms, lengths = zip(*batch)
    padded = pad_sequence(waveforms, batch_first=True)  # (B, T)
    lengths_tensor = torch.stack(lengths, dim=0)
    return padded, lengths_tensor


@pytest.fixture()
def audio_files(test_data_dir):
    """
    Returns audio arrays + sample rate + filepaths for testing.
    """

    audio_file1 = os.path.join(test_data_dir, "an4_speaker", "an4", "wav", "an4_clstk", "fash", "an251-fash-b.wav")
    audio_file2 = os.path.join(test_data_dir, "an4_speaker", "an4", "wav", "an4_clstk", "ffmm", "cen1-ffmm-b.wav")

    audio1, sample_rate1 = sf.read(audio_file1, dtype='float32', always_2d=False)
    audio2, sample_rate2 = sf.read(audio_file2, dtype='float32', always_2d=False)
    assert int(sample_rate1) == int(sample_rate2)

    return audio1, audio2, int(sample_rate1), audio_file1, audio_file2


class DiarizableDummy(DummyModel, SpkDiarizationMixin):
    def _diarize_on_begin(self, audio, diarcfg: DiarizeConfig):
        super()._diarize_on_begin(audio, diarcfg)
        self.flag_begin = True

    def _setup_diarize_dataloader(self, config: Dict) -> DataLoader:
        if "manifest_filepath" in config:
            filepaths: List[str] = []
            with open(config["manifest_filepath"], "r", encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    filepaths.append(json.loads(line)["audio_filepath"])
        else:
            filepaths = list(config["paths2audio_files"])

        return DataLoader(
            dataset=AudioPathDataset(filepaths),
            batch_size=int(config.get("batch_size", 1)),
            num_workers=int(config.get("num_workers", 0)),
            pin_memory=False,
            drop_last=False,
            collate_fn=collate,
        )

    def _diarize_forward(self, batch: Any):
        """
        Real inference step for diarization tests.

        The dataloader yields `(padded_waveforms, lengths)` where:
        - padded_waveforms: (B, T)
        - lengths: (B,)

        We compute a masked mean per sample -> (B, 1) and run it through the model.
        """
        if not isinstance(batch, (tuple, list)) or len(batch) != 2:
            raise TypeError(f"Expected batch=(waveforms, lengths), got: {type(batch)}")

        waveforms, lengths = batch
        if waveforms.dim() != 2:
            raise ValueError(f"Expected waveforms of shape (B, T), got {tuple(waveforms.shape)}")
        if lengths.dim() != 1:
            raise ValueError(f"Expected lengths of shape (B,), got {tuple(lengths.shape)}")

        # Masked mean pooling over time dimension.
        _, T = waveforms.shape
        device = waveforms.device
        t = torch.arange(T, device=device).unsqueeze(0)  # (1, T)
        mask = (t < lengths.unsqueeze(1)).to(waveforms.dtype)  # (B, T)
        denom = lengths.to(waveforms.dtype).clamp_min(1.0).unsqueeze(1)  # (B, 1)
        pooled = (waveforms * mask).sum(dim=1, keepdim=True) / denom  # (B, 1)

        preds = self(pooled)  # (B, 1)
        return preds

    def _diarize_output_processing(self, outputs, uniq_ids, diarcfg: DiarizeConfig):
        self.execution_count += 1
        # Ensure "one scalar per input sample".
        outputs = outputs.detach().cpu().view(outputs.shape[0], -1).mean(dim=1)
        return [float(x) for x in outputs]


@pytest.fixture()
def dummy_model():
    return DiarizableDummy()


class TestSpkDiarizationMixin:
    pytestmark = pytest.mark.with_downloads()

    @pytest.mark.unit
    def test_constructor_non_instance(self):
        model = DummyModel()
        assert not isinstance(model, SpkDiarizationMixin)
        assert not hasattr(model, 'diarize')

    @pytest.mark.unit
    def test_diarize_wav_path_single(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        _, _, _, audio_file1, _ = audio_files
        outputs = dummy_model.diarize(audio_file1, batch_size=1)
        assert isinstance(outputs, list)
        assert len(outputs) == 1
        assert outputs[0] > 0

    @pytest.mark.unit
    def test_diarize_wav_path_list(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        _, _, _, audio_file1, audio_file2 = audio_files
        outputs = dummy_model.diarize([audio_file1, audio_file2], batch_size=1)
        assert isinstance(outputs, list)
        assert len(outputs) == 2
        assert outputs[0] > 0
        assert outputs[1] > 0

    @pytest.mark.unit
    def test_diarize_manifest_jsonl_path(self, dummy_model, audio_files, tmp_path: Path):
        dummy_model = dummy_model.eval()
        _, _, _, audio_file1, audio_file2 = audio_files
        manifest_path = tmp_path / "manifest.jsonl"
        with manifest_path.open("w", encoding="utf-8") as f:
            for audio_file in (audio_file1, audio_file2):
                f.write(json.dumps({"audio_filepath": audio_file, "offset": 0, "duration": None, "text": "-"}) + "\n")
        outputs = dummy_model.diarize(str(manifest_path), batch_size=1)
        assert isinstance(outputs, list)
        assert len(outputs) == 2

    @pytest.mark.unit
    def test_diarize_numpy_single_requires_sample_rate(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        audio1, _, _, _, _ = audio_files

        # Check if it raises an error without sample rate when using a single numpy variable input
        with pytest.raises(ValueError):
            _ = dummy_model.diarize(audio=audio1, batch_size=1)

        # Set sample rate and check if it works
        sample_rate = 16000
        outputs = dummy_model.diarize(audio1, batch_size=1, sample_rate=sample_rate)
        assert isinstance(outputs, list)
        assert len(outputs) == 1
        assert outputs[0] > 0

    @pytest.mark.unit
    def test_diarize_numpy_list_requires_sample_rate(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        audio1, audio2, _, _, _ = audio_files
        numpy_audio_list = [audio1, audio2]
        # Check if it raises an error without sample rate when using numpy list input
        with pytest.raises(ValueError):
            _ = dummy_model.diarize(audio=numpy_audio_list, batch_size=2)

        # Set sample rate and check if it works
        sample_rate = 16000
        outputs = dummy_model.diarize(audio=numpy_audio_list, batch_size=2, sample_rate=sample_rate)
        assert isinstance(outputs, list)
        assert len(outputs) == 2
        assert outputs[0] > 0
        assert outputs[1] > 0

    @pytest.mark.unit
    def test_diarize_numpy_single(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        audio1, _, sample_rate, _, _ = audio_files
        outputs = dummy_model.diarize(audio1, batch_size=1, sample_rate=sample_rate)
        assert isinstance(outputs, list)
        assert len(outputs) == 1
        assert outputs[0] > 0

    @pytest.mark.unit
    def test_diarize_numpy_list(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        audio1, audio2, sample_rate, _, _ = audio_files
        outputs = dummy_model.diarize([audio1, audio2], batch_size=1, sample_rate=sample_rate)
        assert isinstance(outputs, list)
        assert len(outputs) == 2
        assert outputs[0] > 0
        assert outputs[1] > 0

    @pytest.mark.unit
    def test_diarize_numpy_list_but_no_sample_rate(self, dummy_model, audio_files):
        dummy_model = dummy_model.eval()
        # Numpy audio inputs require sample_rate; the mixin should raise with a clear message.
        with pytest.raises(
            ValueError, match=r"Sample rate is not set\. Numpy audio inputs require sample_rate to be set\."
        ):
            _ = dummy_model.diarize(audio_files, batch_size=1)

    @pytest.mark.unit
    def test_transribe_override_config_incorrect(self, dummy_model, audio_files):
        # Not subclassing DiarizeConfig
        @dataclass
        class OverrideConfig:
            batch_size: int = 1
            output_type: str = 'dict'

        dummy_model = dummy_model.eval()

        audio1, _, _, _, _ = audio_files
        override_cfg = OverrideConfig(batch_size=1, output_type='dict')
        with pytest.raises(ValueError):
            _ = dummy_model.diarize(audio1, override_config=override_cfg)

    @pytest.mark.unit
    def test_transribe_override_config_correct(self, dummy_model, audio_files):
        @dataclass
        class OverrideConfig(DiarizeConfig):
            output_type: str = 'dict'
            verbose: bool = False

        dummy_model = dummy_model.eval()
        audio1, _, sample_rate, _, _ = audio_files
        override_cfg = OverrideConfig(batch_size=1, output_type='list', sample_rate=sample_rate)
        outputs = dummy_model.diarize(audio1, override_config=override_cfg)

        assert isinstance(outputs, list)
        assert len(outputs) == 1



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List, Literal, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from transformers import AutoModel, AutoTokenizer
from transformers.cache_utils import Cache
from transformers.modeling_attn_mask_utils import _prepare_4d_attention_mask
from transformers.modeling_outputs import SequenceClassifierOutputWithPast
from transformers.models.llama.configuration_llama import LlamaConfig
from transformers.models.llama.modeling_llama import LlamaForSequenceClassification, LlamaModel
from transformers.utils import logging

logger = logging.get_logger(__name__)


def pool(last_hidden_states: Tensor, attention_mask: Tensor, pool_type: str) -> Tensor:
    """Pooling on last_hidden_states without pad tokens."""
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)

    if pool_type == "avg":
        emb = last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]
    elif pool_type == "weighted_avg":
        emb = last_hidden.sum(dim=1)
    elif pool_type == "cls":
        emb = last_hidden[:, 0]
    elif pool_type == "last":
        left_padding = attention_mask[:, -1].sum() == attention_mask.shape[0]
        if left_padding:
            emb = last_hidden[:, -1]
        else:
            sequence_lengths = attention_mask.sum(dim=1) - 1
            batch_size = last_hidden.shape[0]
            emb = last_hidden[torch.arange(batch_size, device=last_hidden.device), sequence_lengths]
    else:
        raise ValueError(f"pool_type {pool_type} not supported")

    return emb


class LlamaBidirectionalConfig(LlamaConfig):
    """LLamaBidirectionalConfig for LlamaBidirectionalModel."""

    model_type = "llama_bidirec"

    def __init__(
        self,
        pooling="avg",
        temperature=1.0,
        **kwargs,
    ):
        self.pooling = pooling
        self.temperature = temperature
        super().__init__(
            **kwargs,
        )


class LlamaBidirectionalModel(LlamaModel):
    """LlamaBidirectionalModel.
    Attention has been adjusted to bidirectional.
    """

    config_class = LlamaBidirectionalConfig

    def __init__(self, config: LlamaConfig):
        super().__init__(config)
        for layer in self.layers:
            layer.self_attn.is_causal = False

    def _update_causal_mask(
        self,
        attention_mask: torch.Tensor,
        input_tensor: torch.Tensor,
        cache_position: torch.Tensor,
        past_key_values: Cache,
        output_attentions: bool,
    ):
        # Generates bi-directional attention.
        causal_mask = _prepare_4d_attention_mask(attention_mask, input_tensor.dtype)
        return causal_mask


class LlamaBidirectionalForSequenceClassification(LlamaForSequenceClassification):
    """The LLaMa Model transformer with a sequence classification head on top (linear layer)."""

    config_class = LlamaBidirectionalConfig

    def __init__(self, config):
        super().__init__(config)
        # Releasing the parameters of LlamaModel
        # created by parent LlamaForSequenceClassification
        del self.model

        self.model = LlamaBidirectionalModel(config)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Union[Cache, List[torch.FloatTensor]]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, SequenceClassifierOutputWithPast]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]

        pooled_hidden_states = pool(
            last_hidden_states=hidden_states,
            attention_mask=attention_mask,
            pool_type=self.config.pooling,
        )

        pooled_logits = self.score(pooled_hidden_states)
        pooled_logits = pooled_logits / self.config.temperature

        loss = None
        if labels is not None:
            labels = labels.to(pooled_logits.device)
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(pooled_logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(pooled_logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(pooled_logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(pooled_logits, labels)
        if not return_dict:
            output = (pooled_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=pooled_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )


AutoModel.register(LlamaBidirectionalConfig, LlamaBidirectionalModel)  # required for transformers >= 4.57


class LlamaBidirectionalHFAdapter(torch.nn.Module):
    """Wraps a Text embedding model with pooling and normalization."""

    def __init__(
        self,
        model: torch.nn.Module,
        normalize: bool,
        pooling_module: torch.nn.Module,
    ) -> None:
        super().__init__()
        self.model = model
        self.normalize = normalize
        self.pooling_module = pooling_module

    @property
    def device(self) -> torch.device:
        """Returns the device"""

        return self.model.device

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: Optional[torch.Tensor] = None,
        dimensions: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Inference for the adapted Llama model"""

        inputs = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
        }
        if token_type_ids is not None:
            inputs["token_type_ids"] = token_type_ids
        outputs = self.model(**inputs)
        hidden_states = outputs["last_hidden_state"].to(torch.float32)
        embeddings = self.pooling_module(hidden_states, inputs["attention_mask"])

        if dimensions is not None:
            if not torch.all(dimensions > 0):
                raise ValueError("Dimensions must be positive")

            fill_value = torch.tensor(float("-inf"), dtype=embeddings.dtype, device=embeddings.device)

            clipped_dimensions = torch.clamp(dimensions, max=int(embeddings.shape[1]))

            embeddings = embeddings.masked_fill(
                torch.arange(embeddings.shape[1], device=embeddings.device) >= clipped_dimensions.unsqueeze(-1),
                fill_value,
            )[:, : dimensions.max()]

        if self.normalize:
            embeddings = F.normalize(embeddings, p=2, dim=1)

        return embeddings


class Pooling(torch.nn.Module):
    """Pooling layer for the adapter."""

    def __init__(self, pooling_mode: str):
        super().__init__()
        self.pooling_mode = pooling_mode

    def forward(self, last_hidden_states: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        """Forward function of the Pooling layer."""

        last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)

        pool_type = self.pooling_mode
        if pool_type == "avg":
            epsilon = 1e-9  # A small value to avoid division by zero
            emb = last_hidden.sum(dim=1) / (attention_mask.sum(dim=1)[..., None] + epsilon)
        elif pool_type == "cls":  # tokenizer padding right
            emb = last_hidden[:, 0]
        elif pool_type == "cls__left":  # tokenizer padding left
            seq_idxs = (1 - attention_mask).sum(dim=1).to(dtype=torch.long)
            batch_size = last_hidden.shape[0]
            batch_idxs = torch.arange(batch_size, device=last_hidden.device)
            emb = last_hidden[batch_idxs, seq_idxs]
        elif pool_type == "last":  # tokenizer padding left
            emb = last_hidden[:, -1]
        elif pool_type == "last__right":  # tokenizer padding right
            sequence_lengths = (attention_mask.sum(dim=1) - 1).to(dtype=torch.long)
            batch_size = last_hidden.shape[0]
            emb = last_hidden[torch.arange(batch_size, device=last_hidden.device), sequence_lengths]
        else:
            raise ValueError(f"pool_type {pool_type} not supported")

        return emb


def get_llama_bidirectional_hf_model(
    model_name_or_path: Union[str, os.PathLike[str]],
    normalize: bool,
    pooling_mode: Optional[Literal["avg", "cls", "last"]] = None,
    torch_dtype: Optional[Union[torch.dtype, str]] = None,
    trust_remote_code: bool = False,
):
    """Returns the adapter for the Llama bidirectional HF model."""

    # check that the tokenizer matches the requirements of the pooling mode
    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code)
    pooling_mode = pooling_mode or "avg"
    if pooling_mode == "last" and tokenizer.padding_side == "right":
        pooling_mode = "last__right"  # type: ignore
    if pooling_mode == "cls" and tokenizer.padding_side == "left":
        pooling_mode = "cls__left"  # type: ignore

    # load the model
    model = AutoModel.from_pretrained(
        model_name_or_path, torch_dtype=torch_dtype, trust_remote_code=trust_remote_code
    ).eval()

    # configure pooling
    pooling_module = Pooling(pooling_mode=pooling_mode)

    # NV-Embed-v1 model has seperate embedding model and a built-in pooling module
    if (
        model.__class__.__name__ == "NVEmbedModel"
        and hasattr(model, "latent_attention_model")
        and hasattr(model, "embedding_model")
    ):
        pooling_module = model.latent_attention_model
        model = model.embedding_model

    adapted_model = LlamaBidirectionalHFAdapter(model=model, normalize=normalize, pooling_module=pooling_module)
    return adapted_model, tokenizer



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import tempfile

import numpy as np
import pytest
import soundfile as sf
import torch

from nemo.collections.asr.data.audio_to_label import AudioToMultiLabelDataset, TarredAudioToClassificationLabelDataset
from nemo.collections.asr.data.feature_to_label import FeatureToLabelDataset, FeatureToSeqSpeakerLabelDataset
from nemo.collections.asr.parts.preprocessing.feature_loader import ExternalFeatureLoader
from nemo.collections.asr.parts.preprocessing.features import WaveformFeaturizer


class TestASRDatasets:
    labels = ["fash", "fbbh", "fclc"]
    unique_labels_in_seq = ['0', '1', '2', '3', "zero", "one", "two", "three"]

    @pytest.mark.unit
    def test_tarred_dataset(self, test_data_dir):
        manifest_path = os.path.abspath(os.path.join(test_data_dir, 'asr/tarred_an4/tarred_audio_manifest.json'))

        # Test braceexpand loading
        tarpath = os.path.abspath(os.path.join(test_data_dir, 'asr/tarred_an4/audio_{0..1}.tar'))
        featurizer = WaveformFeaturizer(sample_rate=16000, int_values=False, augmentor=None)
        ds_braceexpand = TarredAudioToClassificationLabelDataset(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, featurizer=featurizer
        )

        assert len(ds_braceexpand) == 32
        count = 0
        for _ in ds_braceexpand:
            count += 1
        assert count == 32

        # Test loading via list
        tarpath = [os.path.abspath(os.path.join(test_data_dir, f'asr/tarred_an4/audio_{i}.tar')) for i in range(2)]
        ds_list_load = TarredAudioToClassificationLabelDataset(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, featurizer=featurizer
        )
        count = 0
        for _ in ds_list_load:
            count += 1
        assert count == 32

    @pytest.mark.unit
    def test_tarred_dataset_duplicate_name(self, test_data_dir):
        manifest_path = os.path.abspath(
            os.path.join(test_data_dir, 'asr/tarred_an4/tarred_duplicate_audio_manifest.json')
        )

        # Test braceexpand loading
        tarpath = os.path.abspath(os.path.join(test_data_dir, 'asr/tarred_an4/audio_{0..1}.tar'))
        featurizer = WaveformFeaturizer(sample_rate=16000, int_values=False, augmentor=None)
        ds_braceexpand = TarredAudioToClassificationLabelDataset(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, featurizer=featurizer
        )

        assert len(ds_braceexpand) == 6
        count = 0
        for _ in ds_braceexpand:
            count += 1
        assert count == 6

        # Test loading via list
        tarpath = [os.path.abspath(os.path.join(test_data_dir, f'asr/tarred_an4/audio_{i}.tar')) for i in range(2)]
        ds_list_load = TarredAudioToClassificationLabelDataset(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, featurizer=featurizer
        )
        count = 0
        for _ in ds_list_load:
            count += 1
        assert count == 6

    @pytest.mark.unit
    def test_feat_seqlabel_dataset(self, test_data_dir):
        manifest_path = os.path.abspath(os.path.join(test_data_dir, 'asr/feat/emb.json'))
        feature_loader = ExternalFeatureLoader(augmentor=None)
        ds_braceexpand = FeatureToSeqSpeakerLabelDataset(
            manifest_filepath=manifest_path, labels=self.unique_labels_in_seq, feature_loader=feature_loader
        )
        # fmt: off
        correct_label = torch.tensor(
            [0.0, 1.0, 2.0, 2.0, 1.0, 2.0, 1.0, 2.0, 2.0, 1.0, 2.0, 2.0, 3.0, 1.0, 2.0, 2.0, 2.0, 0.0, 2.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 2.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 2.0, 0.0, 2.0, 2.0, 2.0, 1.0, 2.0, 1.0, 1.0, 0.0, 1.0, 1.0, 0.0, 2.0, 1.0, 2.0, 1.0,]
        )
        # fmt: on
        correct_label_length = torch.tensor(50)

        assert ds_braceexpand[0][0].shape == (50, 32)
        assert torch.equal(ds_braceexpand[0][2], correct_label)
        assert torch.equal(ds_braceexpand[0][3], correct_label_length)

        count = 0
        for _ in ds_braceexpand:
            count += 1
        assert count == 2

    @pytest.mark.unit
    def test_feat_label_dataset(self):

        with tempfile.TemporaryDirectory() as tmpdir:
            manifest_path = os.path.join(tmpdir, 'manifest_input.json')
            with open(manifest_path, 'w', encoding='utf-8') as fp:
                for i in range(2):
                    feat_file = os.path.join(tmpdir, f"feat_{i}.pt")
                    torch.save(torch.randn(80, 5), feat_file)
                    entry = {'feature_file': feat_file, 'duration': 100000, 'label': '0'}
                    fp.write(json.dumps(entry) + '\n')

            dataset = FeatureToLabelDataset(manifest_filepath=manifest_path, labels=self.unique_labels_in_seq)

            correct_label = torch.tensor(self.unique_labels_in_seq.index('0'))
            correct_label_length = torch.tensor(1)

            assert dataset[0][0].shape == (80, 5)
            assert torch.equal(dataset[0][2], correct_label)
            assert torch.equal(dataset[0][3], correct_label_length)

            count = 0
            for _ in dataset:
                count += 1
            assert count == 2

    @pytest.mark.unit
    def test_audio_multilabel_dataset(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            manifest_path = os.path.join(tmpdir, 'manifest_input.json')
            with open(manifest_path, 'w', encoding='utf-8') as fp:
                for i in range(2):
                    audio_file = os.path.join(tmpdir, f"audio_{i}.wav")
                    data = np.random.normal(0, 1, 16000 * 10)
                    sf.write(audio_file, data, 16000)
                    entry = {'audio_filepath': audio_file, 'duration': 10, 'label': '0 1 0 1'}
                    fp.write(json.dumps(entry) + '\n')

            dataset = AudioToMultiLabelDataset(manifest_filepath=manifest_path, sample_rate=16000, labels=['0', '1'])

            correct_label = torch.tensor([0, 1, 0, 1])
            correct_label_length = torch.tensor(4)

            assert dataset[0][0].shape == torch.tensor([0.1] * 160000).shape
            assert torch.equal(dataset[0][2], correct_label)
            assert torch.equal(dataset[0][3], correct_label_length)

            count = 0
            for _ in dataset:
                count += 1
            assert count == 2



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Literal, NotRequired, Optional, TypedDict, Union

import torch
from megatron.core.packed_seq_params import PackedSeqParams
from megatron.energon import Sample

from nemo.collections.multimodal.data.energon.config import AudioToken, MultiModalSampleConfig, VideoToken


@dataclass
class AudioSize:
    """Audio size class for audio sample config"""

    length: int
    channel: int


@dataclass
class VideoSize:
    """Video size class for video sample config"""

    frames: int
    height: int
    width: int


@dataclass
class ImageSize:
    """Image size class for image sample config"""

    height: int
    width: int


class MediaDict(TypedDict):
    """Media dictionary class for media sample config"""

    media_type: Literal["audio", "video", "image"]
    media_value: bytes
    offset: NotRequired[float]
    duration: NotRequired[float]


@dataclass
class AVLMEnergonInterleavedSample(Sample):
    """Sequence of interleaved media, (either str for text, MediaDict for an audio, a video or an image)"""

    sequence: List[Union[bytes, str, MediaDict]]


@dataclass
class AVLMEnergonQASample(Sample):
    """Sample class for question answering sample"""

    context: List[str]
    answers: Optional[List[str]] = None
    answer_weights: Optional[torch.Tensor] = None

    audios: Optional[List[MediaDict]] = None
    videos: Optional[List[MediaDict]] = None
    images: Optional[List[MediaDict]] = None


@dataclass
class AVLMSample:
    """
    Sample type for media to text task, extending LlavaNextTextSample to support audio and video data.

    This class adds additional attributes for handling the audio and video raw bytes,
    along with metadata about the tiled images.

    Attributes:
    """

    __key__: str = ''
    tokens: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.long))
    labels: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.long))
    loss_mask: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))
    audios: Optional[torch.tensor] = None
    audio_lengths: Optional[torch.tensor] = None
    videos: Optional[torch.tensor] = None
    video_lengths: Optional[torch.tensor] = None
    num_video_tiles: Optional[torch.tensor] = None
    images: Optional[torch.tensor] = None
    num_image_tiles: Optional[int] = None
    image_sizes: Optional[torch.tensor] = None
    attention_mask: Optional[torch.tensor] = None


@dataclass
class PackedAVLMSample(AVLMSample):
    """Sample type for packed image audio text sample"""

    __restore_key__: tuple[Union[str, int, tuple], ...] = ()
    position_ids: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))
    packed_seq_params: PackedSeqParams = field(default_factory=lambda: PackedSeqParams())


@dataclass
class AVLMRawBatch:
    """
    Batch type for raw media to text samples, supporting audio, image(s).

    This class aggregates multiple `MediaToMediaSample` instances into a batch for processing.
    It includes attributes for managing audio, image and associated metadata for each sample in the batch.

    Attributes:
    """

    __keys__: List[str] = field(default_factory=list)
    tokens: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.long))
    labels: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.long))
    loss_mask: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))
    audios: Optional[torch.tensor] = None
    audio_lengths: Optional[torch.tensor] = None
    videos: Optional[torch.tensor] = None
    video_lengths: Optional[torch.tensor] = None
    num_video_tiles: Optional[torch.tensor] = None
    images: Optional[torch.tensor] = None
    num_image_tiles: Optional[int] = None
    image_sizes: Optional[torch.tensor] = None
    attention_mask: Optional[torch.tensor] = None


@dataclass
class PackedAVLMRawBatch(AVLMRawBatch):
    """Sample type for image text raw batch"""

    position_ids: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))
    packed_seq_params: PackedSeqParams = field(default_factory=lambda: PackedSeqParams())


@dataclass
class AVLMSampleConfig(MultiModalSampleConfig):
    """
    Sample config for AVLM model
    """

    model_id: str = field(default="llava-hf/llava-v1.6-vicuna-7b-hf")

    # audio related sample config
    audio_encoder_config: Optional[dict] = None
    audio_token: AudioToken = field(default_factory=AudioToken)
    audio_sample_rate: int = field(default=16000)
    audio_channel_selector: Optional[Union[int, Iterable[int], Literal["average"]]] = "average"
    audio_waveform_featurizer_int_values: bool = field(default=False)
    # the detailed structure of audio_augmentor can be found at:
    audio_augmentor: Optional[Dict[str, Dict[str, Any]]] = None

    # video related sample config
    video_token: VideoToken = field(default_factory=VideoToken)
    """
    For a single video with multiple video and audio streams
    sequential: the ordering of the stream tokens follows the stream index from the video container's muxer
    video_audio: video streams tokens are before the audio streams tokens 
    audio_video: audio streams tokens are before the video streams tokens
    interleaved_optimal: space the video tokens and the audio tokens as evenly as possible
    """
    audio_video_tokens_concatenate_pattern: Literal[
        "sequential",
        "audio_video",
        "video_audio",
        "interleaved_optimal",
    ] = field(default="sequential")

    # image related sample config
    image_encoder_config: Optional[dict] = None



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import gc
import tempfile
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import torch
from jiwer import wer as word_error_rate
from omegaconf import DictConfig

from nemo.collections.asr.parts.utils.wfst_utils import TW_BREAK, kaldifst_importer

RIVA_DECODER_INSTALLATION_MESSAGE = (
    "riva decoder is not installed or is installed incorrectly.\n"
    "please run `bash scripts/installers/install_riva_decoder.sh` or `pip install riva-asrlib-decoder` to install."
)


def riva_decoder_importer():
    """Import helper function that returns Riva asrlib decoder package or raises ImportError exception."""
    try:
        import riva.asrlib.decoder.python_decoder as riva_decoder
    except (ImportError, ModuleNotFoundError):
        raise ImportError(RIVA_DECODER_INSTALLATION_MESSAGE)
    return riva_decoder


def _riva_config_to_dict(conf: Any) -> Dict[str, Any]:
    """
    Helper function for parsing Riva configs (namely BatchedMappedDecoderCudaConfig) into a dictionary.

    Args:
      conf:
        Inner Riva config.

    Returns:
      Dictionary corresponding to the Riva config.
    """
    result = {}
    for name in conf.__dir__():
        if not name.startswith("__"):
            attribute = getattr(conf, name)
            result[name] = (
                attribute if attribute.__class__.__module__ == 'builtins' else _riva_config_to_dict(attribute)
            )
    return result


def _fill_inner_riva_config_(riva_conf, nemo_conf):
    """
    Helper function for filling Riva configs (namely BatchedMappedDecoderCudaConfig)
    according to the corresponding NeMo config.

    Note: in-place for the first argument.

    Args:
      riva_conf:
        Inner Riva config.

      nemo_conf:
        Corresponding NeMo config.
    """
    for nemo_k, nemo_v in nemo_conf.items():
        if isinstance(nemo_v, DictConfig):
            _fill_inner_riva_config_(getattr(riva_conf, nemo_k), nemo_v)
        else:
            setattr(riva_conf, nemo_k, nemo_v)


class RivaDecoderConfig(DictConfig):
    """
    NeMo config for the RivaGpuWfstDecoder.
    """

    def __init__(self):
        try:
            riva_decoder = riva_decoder_importer()

            config = riva_decoder.BatchedMappedDecoderCudaConfig()
            config.online_opts.lattice_postprocessor_opts.acoustic_scale = 10.0
            config.n_input_per_chunk = 50
            config.online_opts.decoder_opts.default_beam = 20.0
            config.online_opts.decoder_opts.max_active = 10000
            config.online_opts.determinize_lattice = True
            config.online_opts.max_batch_size = 800
            config.online_opts.num_channels = 800
            config.online_opts.frame_shift_seconds = 1  # not actual frame shift
            config.online_opts.lattice_postprocessor_opts.word_ins_penalty = 0.0

            content = _riva_config_to_dict(config)
        except ImportError:
            content = {}
        super().__init__(content)


class WfstNbestUnit(NamedTuple):
    """
    Container for a single RivaGpuWfstDecoder n-best hypothesis.
    """

    words: Tuple[str]
    timesteps: Tuple[int]
    alignment: Tuple[int]
    score: float


class WfstNbestHypothesis:
    """
    Container for the RivaGpuWfstDecoder n-best results represented as a list of WfstNbestUnit objects.
    """

    def __init__(self, raw_hypotheses: Tuple[Tuple[Tuple[str], Tuple[int], Tuple[int], float]]):
        for i, rh in enumerate(raw_hypotheses):
            assert isinstance(rh[0], tuple), f"{rh[0]}"
            assert isinstance(rh[1], tuple), f"{rh[1]}, {rh[0]}"
            assert isinstance(rh[2], tuple), f"{rh[2]}"
            assert isinstance(rh[3], float), f"{rh[3]}"
            assert len(rh[0]) == len(rh[1]) or len(rh[1]) == 0, "words do not match timesteps"

        self._hypotheses = sorted([WfstNbestUnit(*rh) for rh in raw_hypotheses], key=lambda hyp: hyp.score)
        self._shape0 = len(self._hypotheses)
        self._shape1 = [len(h.words) for h in self._hypotheses]
        self._has_timesteps = len(self._hypotheses[0].timesteps) > 0
        self._has_alignment = len(self._hypotheses[0].alignment) > 0

    def __iter__(self):
        yield from self._hypotheses

    def __getitem__(self, index):
        return self._hypotheses[index]

    def __len__(self):
        return self.shape0

    def replace_unit_(
        self, index: int, new_unit: Union[WfstNbestUnit, Tuple[Tuple[str], Tuple[int], Tuple[int], float]]
    ):
        """
        Replaces a WfstNbestUnit by index.

        Note: in-place operation.

        Args:
          index:
            Index of the unit to be replaced.

          new_unit:
            Replacement unit.
        """
        assert 0 <= index < self.shape0
        assert (
            self.has_timesteps
            and len(new_unit[0]) == len(new_unit[1])
            or not self.has_timesteps
            and len(new_unit[1]) == 0
        )
        assert (
            index == 0
            and (len(self._hypotheses) == 1 or new_unit[3] <= self._hypotheses[index + 1].score)
            or index == self.shape0 - 1
            and self._hypotheses[index - 1].score <= new_unit[3]
            or self._hypotheses[index - 1].score <= new_unit[3] <= self._hypotheses[index + 1].score
        )

        if not isinstance(new_unit, WfstNbestUnit):
            new_unit = WfstNbestUnit(*new_unit)
        self._hypotheses[index] = new_unit
        self._shape1[index] = len(new_unit.words)

    @property
    def shape0(self):
        return self._shape0

    @property
    def shape1(self):
        return self._shape1

    @property
    def has_timesteps(self):
        return self._has_timesteps

    @property
    def has_alignment(self):
        return self._has_alignment


def collapse_tokenword_hypotheses(
    hypotheses: List[WfstNbestHypothesis], tokenword_disambig_str: str
) -> List[WfstNbestHypothesis]:
    """
    Searches for tokenwords in the input hypotheses and collapses them into words.

    Args:
      hypotheses:
        List of input WfstNbestHypothesis.

      tokenword_disambig_str:
        Tokenword disambiguation symbol (e.g. `#1`).

    Returns:
      List of WfstNbestHypothesis.
    """
    new_hypotheses = copy.deepcopy(hypotheses)
    for hyp in new_hypotheses:
        for k, h_unit in enumerate(hyp):
            twds_list = []
            for i, word in enumerate(h_unit.words):
                if word == tokenword_disambig_str:
                    twds_list.append(i)
            if len(twds_list) > 0:
                # a rare case when the recognition stopped before completing the tokenword
                old_words = list(h_unit.words)
                old_timesteps = list(h_unit.timesteps)
                words_len = len(old_words)
                if len(twds_list) % 2 == 1:
                    twds_list.append(words_len)
                new_words, new_timesteps = [], []
                j_prev = 0
                for i, j in zip(twds_list[::2], twds_list[1::2]):
                    new_words += old_words[j_prev:i]
                    # drop tokenword disambig -> remove token disanbig suffix -> remove word begin mark
                    new_word = "".join(old_words[i + 1 : j]).replace(f"{TW_BREAK}{tokenword_disambig_str}", "")[1:]
                    new_words.append(new_word)
                    new_timesteps += old_timesteps[j_prev:i] + [
                        old_timesteps[i],
                    ]
                    j_prev = j + 1
                if j_prev < words_len:
                    new_words += old_words[j_prev:words_len]
                    new_timesteps += old_timesteps[j_prev:words_len]
                hyp.replace_unit_(k, (tuple(new_words), tuple(new_timesteps), h_unit.alignment, h_unit.score))
    return new_hypotheses


class AbstractWFSTDecoder(ABC):
    """
    Used for performing WFST decoding of the logprobs.

    Args:
      lm_fst:
        Language model WFST.

      decoding_mode:
        Decoding mode. E.g. `nbest`.

      beam_size:
        Beam width (float) for the WFST decoding.

      config:
        Decoder config.

      tokenword_disambig_id:
        Tokenword disambiguation index. Set to -1 to disable the tokenword mode.

      lm_weight:
        Language model weight in decoding.
    """

    def __init__(
        self,
        lm_fst: Any,
        decoding_mode: str,
        beam_size: float,
        config: Optional[Any],
        tokenword_disambig_id: int = -1,
        lm_weight: float = 1.0,
    ):
        self._lm_fst = lm_fst
        self._beam_size = beam_size
        self._tokenword_disambig_id = tokenword_disambig_id
        self._open_vocabulary_decoding = self._tokenword_disambig_id >= 0
        self._lm_weight = lm_weight
        self._id2word, self._word2id = None, None
        self._id2token, self._token2id = None, None
        self._decoding_mode, self._config, self._decoder = None, None, None

        self._set_decoding_mode(decoding_mode)
        self._set_decoder_config(config)
        self._init_decoder()

    @abstractmethod
    def _set_decoder_config(self, config: Optional[Any] = None):
        pass

    @abstractmethod
    def _set_decoding_mode(self, decoding_mode: str):
        pass

    @abstractmethod
    def _init_decoder(self):
        pass

    @property
    def decoding_mode(self):
        return self._decoding_mode

    @decoding_mode.setter
    def decoding_mode(self, value: str):
        self._decoding_mode_setter(value)

    @abstractmethod
    def _decoding_mode_setter(self, value: str):
        pass

    @property
    def beam_size(self):
        return self._beam_size

    @beam_size.setter
    def beam_size(self, value: float):
        self._beam_size_setter(value)

    @abstractmethod
    def _beam_size_setter(self, value: float):
        pass

    @property
    def lm_weight(self):
        return self._lm_weight

    @lm_weight.setter
    def lm_weight(self, value: float):
        self._lm_weight_setter(value)

    @abstractmethod
    def _lm_weight_setter(self, value: float):
        pass

    @property
    def tokenword_disambig_id(self):
        return self._tokenword_disambig_id

    @property
    def open_vocabulary_decoding(self):
        return self._open_vocabulary_decoding

    @abstractmethod
    def decode(self, log_probs: torch.Tensor, log_probs_length: torch.Tensor) -> List[Any]:
        """
        Decodes logprobs into recognition hypotheses.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

        Returns:
          List of recognition hypotheses.
        """
        pass

    @abstractmethod
    def _post_decode(self, hypotheses: List[Any]) -> List[Any]:
        """
        Does various post-processing of the recognition hypotheses.

        Args:
          hypotheses:
            List of recognition hypotheses.

        Returns:
          List of processed recognition hypotheses.
        """
        pass

    @abstractmethod
    def calibrate_lm_weight(
        self, log_probs: torch.Tensor, log_probs_length: torch.Tensor, reference_texts: List[str]
    ) -> Tuple[float, float]:
        """
        Calibrates LM weight to achieve the best WER for given logprob-text pairs.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

          reference_texts:
            List of reference word sequences.

        Returns:
          Pair of (best_lm_weight, best_wer).
        """
        pass

    @abstractmethod
    def calculate_oracle_wer(
        self, log_probs: torch.Tensor, log_probs_length: torch.Tensor, reference_texts: List[str]
    ) -> Tuple[float, List[float]]:
        """
        Calculates the oracle (the best possible WER for given logprob-text pairs.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

          reference_texts:
            List of reference word sequences.

        Returns:
          Pair of (oracle_wer, oracle_wer_per_utterance).
        """
        pass


class RivaGpuWfstDecoder(AbstractWFSTDecoder):
    """
    Used for performing WFST decoding of the logprobs with the Riva WFST decoder.

    Args:
      lm_fst:
        Kaldi-type language model WFST or its path.

      decoding_mode:
        Decoding mode. Choices: `nbest`, `mbr`, `lattice`.

      beam_size:
        Beam width (float) for the WFST decoding.

      config:
        Riva Decoder config.

      tokenword_disambig_id:
        Tokenword disambiguation index. Set to -1 to disable the tokenword mode.

      lm_weight:
        Language model weight in decoding.

      nbest_size:
        N-best size for decoding_mode == `nbest`
    """

    def __init__(
        self,
        lm_fst: Union['kaldifst.StdFst', Path, str],
        decoding_mode: str = 'mbr',
        beam_size: float = 10.0,
        config: Optional['RivaDecoderConfig'] = None,
        tokenword_disambig_id: int = -1,
        lm_weight: float = 1.0,
        nbest_size: int = 1,
    ):
        self._nbest_size = nbest_size
        self._load_word_lattice = None
        super().__init__(lm_fst, decoding_mode, beam_size, config, tokenword_disambig_id, lm_weight)

    def _set_decoder_config(self, config: Optional['RivaDecoderConfig'] = None):
        if config is None or len(config) == 0:
            config = RivaDecoderConfig()
        if not hasattr(config, "online_opts"):
            # most likely empty config
            # call importer to raise the exception + installation message
            riva_decoder_importer()
            # just in case
            raise RuntimeError("Unexpected config error. Please debug manually.")
        config.online_opts.decoder_opts.lattice_beam = self._beam_size
        config.online_opts.lattice_postprocessor_opts.lm_scale = (
            self._lm_weight * config.online_opts.lattice_postprocessor_opts.acoustic_scale
        )
        config.online_opts.lattice_postprocessor_opts.nbest = self._nbest_size
        self._config = config

    def _init_decoder(self):

        # use importers instead of direct import to possibly get an installation message
        kaldifst = kaldifst_importer()
        riva_decoder = riva_decoder_importer()

        from nemo.collections.asr.parts.utils.wfst_utils import load_word_lattice

        self._load_word_lattice = load_word_lattice
        # BatchedMappedDecoderCuda supports filepaths only
        # TODO: fix when possible
        lm_fst = self._lm_fst
        tmp_fst = None
        tmp_fst_file = None
        if isinstance(lm_fst, (Path, str)):
            # We only read lm_fst to extract words.txt and num_tokens_with_blank
            tmp_fst = kaldifst.StdVectorFst.read(lm_fst)
        elif isinstance(lm_fst, (kaldifst.StdVectorFst, kaldifst.StdConstFst)):
            tmp_fst = lm_fst
            tmp_fst_file = tempfile.NamedTemporaryFile(mode='w+t')
            tmp_fst.write(tmp_fst_file.name)
            lm_fst = tmp_fst_file.name
        else:
            raise ValueError(f"Unsupported lm_fst type: {type(lm_fst)}")

        # we assume that lm_fst has at least one disambig after real tokens
        num_tokens_with_blank = tmp_fst.input_symbols.find('#0') - 1
        if self._id2word is None:
            self._id2word = {
                int(line.split("\t")[1]): line.split("\t")[0]
                for line in str(tmp_fst.output_symbols).strip().split("\n")
            }
            word2id = self._id2word.__class__(map(reversed, self._id2word.items()))
            word_unk_id = word2id["<unk>"]
            self._word2id = defaultdict(lambda: word_unk_id)
            for k, v in word2id.items():
                self._word2id[k] = v
        if self._id2token is None:
            self._id2token = {
                int(line.split("\t")[1]): line.split("\t")[0]
                for line in str(tmp_fst.input_symbols).strip().split("\n")
            }
            token2id = self._id2token.__class__(map(reversed, self._id2token.items()))
            token_unk_id = token2id["<unk>"]
            self._token2id = defaultdict(lambda: token_unk_id)
            for k, v in token2id.items():
                self._token2id[k] = v
        with tempfile.NamedTemporaryFile(mode='w+t') as words_tmp:
            tmp_fst.output_symbols.write_text(words_tmp.name)
            config = riva_decoder.BatchedMappedDecoderCudaConfig()
            _fill_inner_riva_config_(config, self._config)
            self._decoder = riva_decoder.BatchedMappedDecoderCuda(
                config, lm_fst, words_tmp.name, num_tokens_with_blank
            )
        if tmp_fst_file:
            tmp_fst_file.close()

    def _set_decoding_mode(self, decoding_mode: str):
        if decoding_mode == 'nbest':
            self._decode = self._decode_nbest
        elif decoding_mode == 'mbr':
            self._decode = self._decode_mbr
        elif decoding_mode == 'lattice':
            self._decode = self._decode_lattice
        else:
            raise ValueError(f"Unsupported mode: {decoding_mode}")
        self._decoding_mode = decoding_mode

    def _beam_size_setter(self, value: float):
        if self._beam_size != value:
            self._release_gpu_memory()
            self._config.online_opts.decoder_opts.lattice_beam = value
            self._init_decoder()
            self._beam_size = value

    def _lm_weight_setter(self, value: float):
        if self._lm_weight != value:
            self._release_gpu_memory()
            self._config.online_opts.lattice_postprocessor_opts.lm_scale = (
                value * self._config.online_opts.lattice_postprocessor_opts.acoustic_scale
            )
            self._init_decoder()
            self._lm_weight = value

    def _decoding_mode_setter(self, value: str):
        if self._decoding_mode != value:
            self._set_decoding_mode(value)

    @property
    def nbest_size(self):
        return self._nbest_size

    @nbest_size.setter
    def nbest_size(self, value: float):
        self._nbest_size_setter(value)

    def _nbest_size_setter(self, value: float):
        if self._nbest_size != value:
            self._release_gpu_memory()
            self._config.online_opts.lattice_postprocessor_opts.nbest = value
            self._init_decoder()
            self._nbest_size = value

    def _decode_nbest(
        self, log_probs: torch.Tensor, log_probs_length: torch.Tensor
    ) -> List[WfstNbestHypothesis]:  # words, timesteps, alignment, score
        """
        Decodes logprobs into recognition hypotheses via the N-best decoding decoding.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

        Returns:
          List of WfstNbestHypothesis with empty alignment and trivial score.
        """
        hypotheses_nbest = self._decoder.decode_nbest(log_probs, log_probs_length)
        hypotheses = []
        for nh in hypotheses_nbest:
            nbest_container = []
            for h in nh:
                words, timesteps = [], []
                for w, t in zip(h.words, h.word_start_times_seconds):
                    if w != 0:
                        words.append(self._id2word[w])
                        timesteps.append(int(t))
                alignment = [ilabel - 1 for ilabel in h.ilabels]
                score = h.score
                nbest_container.append(tuple([tuple(words), tuple(timesteps), tuple(alignment), score]))
            hypotheses.append(WfstNbestHypothesis(tuple(nbest_container)))
        return hypotheses

    def _decode_mbr(self, log_probs: torch.Tensor, log_probs_length: torch.Tensor) -> List[WfstNbestHypothesis]:
        """
        Decodes logprobs into recognition hypotheses via the Minimum Bayes Risk (MBR) decoding.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

        Returns:
          List of WfstNbestHypothesis with empty alignment and trivial score.
        """
        hypotheses_mbr = self._decoder.decode_mbr(log_probs, log_probs_length)
        hypotheses = []
        for h in hypotheses_mbr:
            words, timesteps = [], []
            for e in h:
                words.append(e[0])
                timesteps.append(int(e[1]))
            hypotheses.append(WfstNbestHypothesis(tuple([tuple([tuple(words), tuple(timesteps), tuple(), 0.0])])))
        return hypotheses

    def _decode_lattice(self, log_probs: torch.Tensor, log_probs_length: torch.Tensor) -> List['KaldiWordLattice']:
        """
        Decodes logprobs into kaldi-type lattices.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

        Returns:
          List of KaldiWordLattice.
        """
        with tempfile.NamedTemporaryFile() as tmp_lat:
            tmp_lat_name = f"{tmp_lat.name}.lats"
            self._decoder.decode_write_lattice(
                log_probs, log_probs_length, [str(i) for i in range(len(log_probs))], f"ark,t:{tmp_lat_name}"
            )
            hypotheses_lattice = self._load_word_lattice(
                tmp_lat_name, self._id2word, self._id2word
            )  # input and output token ids are the same
            hypotheses = [hypotheses_lattice[str(i)] for i in range(len(log_probs))]
        return hypotheses

    def decode(
        self, log_probs: torch.Tensor, log_probs_length: torch.Tensor
    ) -> Union[List[WfstNbestHypothesis], List['KaldiWordLattice']]:
        """
        Decodes logprobs into recognition hypotheses.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

        Returns:
          List of recognition hypotheses.
        """
        log_probs = log_probs.contiguous()
        log_probs_length = log_probs_length.to(torch.long).to('cpu').contiguous()
        hypotheses = self._decode(log_probs, log_probs_length)
        hypotheses = self._post_decode(hypotheses)
        return hypotheses

    def _post_decode(
        self, hypotheses: Union[List[WfstNbestHypothesis], List['KaldiWordLattice']]
    ) -> Union[List[WfstNbestHypothesis], List['KaldiWordLattice']]:
        """
        Does various post-processing of the recognition hypotheses.

        Args:
          hypotheses:
            List of recognition hypotheses.

        Returns:
          List of processed recognition hypotheses.
        """
        if self._open_vocabulary_decoding and self._decoding_mode in ('nbest', 'mbr'):
            return collapse_tokenword_hypotheses(hypotheses, self._id2word[self._tokenword_disambig_id])
        else:
            return hypotheses

    def calibrate_lm_weight(
        self, log_probs: torch.Tensor, log_probs_length: torch.Tensor, reference_texts: List[str]
    ) -> Tuple[float, float]:
        """
        Calibrates LM weight to achieve the best WER for given logprob-text pairs.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

          reference_texts:
            List of reference word sequences.

        Returns:
          Pair of (best_lm_weight, best_wer).
        """
        assert len(log_probs) == len(reference_texts)
        decoding_mode_backup = self.decoding_mode
        lm_weight_backup = self.lm_weight
        self.decoding_mode = "mbr"
        best_lm_weight, best_wer = -1.0, float('inf')
        for lm_weight in range(1, 21):  # enough for most cases
            self.lm_weight = lm_weight / 10
            hypotheses = self.decode(log_probs, log_probs_length)
            wer = word_error_rate([" ".join(h[0].words) for h in hypotheses], reference_texts)
            print(lm_weight, wer)
            if wer < best_wer:
                best_lm_weight, best_wer = self.lm_weight, wer
        self.decoding_mode = decoding_mode_backup
        self.lm_weight = lm_weight_backup
        return best_lm_weight, best_wer

    def calculate_oracle_wer(
        self, log_probs: torch.Tensor, log_probs_length: torch.Tensor, reference_texts: List[str]
    ) -> Tuple[float, List[float]]:
        """
        Calculates the oracle (the best possible WER for given logprob-text pairs.

        Args:
          log_probs:
            A torch.Tensor of the predicted log-probabilities of shape [Batch, Time, Vocabulary].

          log_probs_length:
            A torch.Tensor of length `Batch` which contains the lengths of the log_probs elements.

          reference_texts:
            List of reference word sequences.

        Returns:
          Pair of (oracle_wer, oracle_wer_per_utterance).
        """
        if self._open_vocabulary_decoding:
            raise NotImplementedError
        assert len(log_probs) == len(reference_texts)
        decoding_mode_backup = self.decoding_mode
        self.decoding_mode = "lattice"
        lattices = self.decode(log_probs, log_probs_length)
        scores, counts, wer_per_utt = [], [], []
        for lattice, text in zip(lattices, reference_texts):
            word_ids = [self._word2id[w] for w in text.strip().split()]
            counts.append(len(word_ids) if word_ids else 1)
            scores.append(lattice.edit_distance(word_ids))
            wer_per_utt.append(scores[-1] / counts[-1])
        self.decoding_mode = decoding_mode_backup
        return sum(scores) / sum(counts), wer_per_utt

    def _release_gpu_memory(self):
        """
        Forces freeing of GPU memory by deleting the Riva decoder object.
        """
        try:
            del self._decoder
        except Exception:
            # apparently self._decoder was previously deleted, do nothing
            pass
        gc.collect()

    def __del__(self):
        self._release_gpu_memory()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import tempfile
from unittest import mock

import pytest
import torch
import torch.cuda
from omegaconf import DictConfig

from nemo.collections.asr.data.audio_to_diar_label_lhotse import LhotseAudioToSpeechE2ESpkDiarDataset
from nemo.collections.common.data.lhotse import get_lhotse_dataloader_from_config


def get_train_ds_config(manifest_filepath, batch_size, num_workers) -> DictConfig:
    return DictConfig(
        {
            'manifest_filepath': manifest_filepath,
            'sample_rate': 16000,
            'num_spks': 4,
            'session_len_sec': 90,
            'soft_label_thres': 0.5,
            'soft_targets': False,
            'labels': None,
            'batch_size': batch_size,
            'shuffle': True,
            'num_workers': num_workers,
            'validation_mode': False,
            'use_lhotse': True,
            'use_bucketing': True,
            'num_buckets': 10,
            'bucket_duration_bins': [10, 20, 30, 40, 50, 60, 70, 80, 90],
            'pin_memory': True,
            'min_duration': 80,
            'max_duration': 90,
            'batch_duration': 400,
            'quadratic_duration': 1200,
            'bucket_buffer_size': 20000,
            'shuffle_buffer_size': 10000,
            'window_stride': 0.01,
            'subsampling_factor': 8,
        }
    )


def get_validation_ds_config(manifest_filepath, batch_size, num_workers) -> DictConfig:
    return DictConfig(
        {
            'manifest_filepath': manifest_filepath,
            'is_tarred': False,
            'tarred_audio_filepaths': None,
            'sample_rate': 16000,
            'num_spks': 4,
            'session_len_sec': 90,
            'soft_label_thres': 0.5,
            'soft_targets': False,
            'labels': None,
            'batch_size': batch_size,
            'shuffle': False,
            'seq_eval_mode': True,
            'num_workers': num_workers,
            'validation_mode': True,
            'use_lhotse': False,
            'use_bucketing': False,
            'drop_last': False,
            'pin_memory': True,
            'window_stride': 0.01,
            'subsampling_factor': 8,
        }
    )


def get_test_ds_config(manifest_filepath, batch_size, num_workers) -> DictConfig:
    return DictConfig(
        {
            'manifest_filepath': manifest_filepath,
            'is_tarred': False,
            'tarred_audio_filepaths': None,
            'sample_rate': 16000,
            'num_spks': 4,
            'session_len_sec': 90,
            'soft_label_thres': 0.5,
            'soft_targets': False,
            'labels': None,
            'batch_size': batch_size,
            'shuffle': False,
            'seq_eval_mode': True,
            'num_workers': num_workers,
            'validation_mode': True,
            'use_lhotse': False,
            'use_bucketing': False,
            'drop_last': False,
            'pin_memory': True,
            'window_stride': 0.01,
            'subsampling_factor': 8,
        }
    )


class TestLhotseAudioToSpeechE2ESpkDiarDataset:

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "batch_size, num_workers, split",
        [
            (4, 8, 'train'),  # Example 1
            (4, 0, 'train'),  # Example 2
            (2, 4, 'validation'),  # Example 3
            (8, 2, 'test'),  # Example 4
        ],
    )
    def test_e2e_speaker_diar_lhotse_dataset(self, test_data_dir, batch_size, num_workers, split):
        manifest_path = os.path.abspath(os.path.join(test_data_dir, 'asr/diarizer/lsm_val.json'))
        num_samples = 8
        device = 'gpu' if torch.cuda.is_available() else 'cpu'
        data_dict_list = []
        with tempfile.NamedTemporaryFile(mode='w', encoding='utf-8') as f:
            with open(manifest_path, 'r', encoding='utf-8') as mfile:
                for ix, line in enumerate(mfile):
                    if ix >= num_samples:
                        break

                    line = line.replace("tests/data/", test_data_dir + "/").replace("\n", "")
                    f.write(f"{line}\n")
                    data_dict = json.loads(line)
                    data_dict_list.append(data_dict)

            f.seek(0)
            config = None
            if split == 'train':
                config = get_train_ds_config(manifest_filepath=f.name, batch_size=batch_size, num_workers=num_workers)
            elif split == 'validation':
                config = get_train_ds_config(manifest_filepath=f.name, batch_size=batch_size, num_workers=num_workers)
            elif split == 'test':
                config = get_test_ds_config(manifest_filepath=f.name, batch_size=batch_size, num_workers=num_workers)

            dataloader_instance = get_lhotse_dataloader_from_config(
                config,
                global_rank=0,
                world_size=1,
                dataset=LhotseAudioToSpeechE2ESpkDiarDataset(cfg=config),
            )

            deviation_thres_rate = 0.01  # 1% deviation allowed
            for batch_index, batch in enumerate(dataloader_instance):
                audio_signals, audio_signal_len, targets, target_lens = batch
                for sample_index in range(audio_signals.shape[0]):
                    dataloader_audio_in_sec = audio_signal_len[sample_index].item()
                    data_dur_in_sec = abs(
                        data_dict_list[batch_size * batch_index + sample_index]['duration'] * config.sample_rate
                        - dataloader_audio_in_sec
                    )
                    assert (
                        data_dur_in_sec <= deviation_thres_rate * dataloader_audio_in_sec
                    ), "Duration deviation exceeds 1%"
                assert not torch.isnan(audio_signals).any(), "audio_signals tensor contains NaN values"
                assert not torch.isnan(audio_signal_len).any(), "audio_signal_len tensor contains NaN values"
                assert not torch.isnan(targets).any(), "targets tensor contains NaN values"
                assert not torch.isnan(target_lens).any(), "target_lens tensor contains NaN values"
